<?php
/************************************************************************
 * MXChange v0.2.1                                    Start: 08/25/2003 *
 * ===============                              Last change: 11/29/2005 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : functions.php                                    *
 * -------------------------------------------------------------------- *
 * Short description : Many non-MySQL functions (also file access)      *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Viele Nicht-MySQL-Funktionen (auch Dateizugriff) *
 * -------------------------------------------------------------------- *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2008 by Roland Haeder                           *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	$INC = substr(dirname(__FILE__), 0, strpos(dirname(__FILE__), "/inc") + 4)."/security.php";
	require($INC);
}

// Check if our config file is writeable or not
function IS_INC_WRITEABLE($inc) {
	// Generate FQFN
	$fqfn = sprintf("%sinc/%s.php", PATH, $inc);

	// Abort by simple test
	if ((FILE_READABLE($fqfn)) && (!is_writeable($fqfn))) {
		return false;
	} // END - if

	// Test if we can append data
	$fp = @fopen($fqfn, 'a');
	if ($inc == "dummy") {
		// Remove dummy file
		@fclose($fp);
		return @unlink($fqfn);
	} else {
		// Close all other files
		return @fclose($fp);
	}
}

// Open a table (you may want to add some header stuff here)
function OPEN_TABLE($PERCENT = "", $CLASS = "", $ALIGN="left", $VALIGN="", $td_only=false) {
	global $table_cnt;

	// Count tables so we can generate CSS classes for every table... :-)
	if (empty($CLASS)) {
		// Class is empty so count one up and create a class
		$table_cnt++; $CLASS = "class".$table_cnt;
	}
	$OUT = "<TABLE border=\"0\" cellspacing=\"0\" cellpadding=\"0\"";

	// Shall I add the classes to TABLE and TD or only to TD?
	if (!$td_only) $OUT .= " class=\"".$CLASS."\"";

	// Width is given
	if (!empty($PERCENT)) $OUT .= " width=\"".$PERCENT."\"";

	// Horizonal align
	if (!empty($ALIGN)) $OUT .=" align=\"".$ALIGN."\"";

	// Vertical align is given
	if (!empty($VALIGN))  $OUT .= " valign=\"".$VALIGN."\"";
	$OUT .= ">\n<TR>\n<TD";
	if (!empty($ALIGN)) $OUT .=" align=\"".$ALIGN."\"";
	$OUT .= " class=\"".$CLASS."\">";
	OUTPUT_HTML($OUT);
}

// Close a table (you may want to add some footer stuff here)
function CLOSE_TABLE($ADD="") {
	OUTPUT_HTML("  </TD>\n</TR>");
	if (!empty($ADD)) OUTPUT_HTML($ADD);
	OUTPUT_HTML("</TABLE>");
}

// Output HTML code directly or "render" it. You addionally switch the new-line character off
function OUTPUT_HTML($HTML, $NEW_LINE = true) {
	// Some global variables
	global $OUTPUT, $footer, $CSS;

	// Do we have HTML-Code here?
	if (!empty($HTML)) {
		// Yes, so we handle it as you have configured
		switch (OUTPUT_MODE)
		{
		case "render":
			// That's why you don't need any \n at the end of your HTML code... :-)
			if (_OB_CACHING == "on") {
				// Output into PHP's internal buffer
				OUTPUT_RAW($HTML);

				// That's why you don't need any \n at the end of your HTML code... :-)
				if ($NEW_LINE) echo "\n";
			} else {
				// Render mode for old or lame servers...
				$OUTPUT .= $HTML;

				// That's why you don't need any \n at the end of your HTML code... :-)
				if ($NEW_LINE) $OUTPUT .= "\n";
			}
			break;

		case "direct":
			// If we are switching from render to direct output rendered code
			if ((!empty($OUTPUT)) && (_OB_CACHING != "on")) { OUTPUT_RAW($OUTPUT); $OUTPUT = ""; }

			// The same as above... ^
			OUTPUT_RAW($HTML);
			if ($NEW_LINE) echo "\n";
			break;

		default:
			// Huh, something goes wrong or maybe you have edited config.php ???
			DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Invalid renderer %s detected.", OUTPUT_MODE));
			MXCHANGE_DIE("<STRONG>".FATAL_ERROR.":</STRONG> ".LANG_NO_RENDER_DIRECT);
			break;
		}
	} elseif ((_OB_CACHING == "on") && ($footer == 1)) {
		// Output cached HTML code
		$OUTPUT = ob_get_contents();

		// Clear output buffer for later output
		ob_end_clean();

		// Send HTTP header
		header("HTTP/1.1 200");

		// Used later
		$now = gmdate('D, d M Y H:i:s') . ' GMT';

		// General headers for no caching
		header("Expired: " . $now); // RFC2616 - Section 14.21
		header("Last-Modified: " . $now);
		header("Cache-Control: no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0"); // HTTP/1.1
		header("Pragma: no-cache"); // HTTP/1.0
		header("Connection: Close");

		// Extension "rewrite" installed?
		if ((EXT_IS_ACTIVE("rewrite")) && (function_exists('REWRITE_LINKS')) && ($CSS != "1") && ($CSS != "-1")) {
			$OUTPUT = REWRITE_LINKS($OUTPUT);
		} // END - if

		// Compile and run finished rendered HTML code
		while (strpos($OUTPUT, '{!') > 0) {
			// Prepare the content and eval() it...
			$newContent = "";
			$eval = "\$newContent = \"".COMPILE_CODE(addslashes($OUTPUT))."\";";
			@eval($eval);

			// Was that eval okay?
			if (empty($newContent)) {
				// Something went wrong!
				die("Evaluation error:<pre>".htmlentities($eval)."</pre>");
			} // END - if
			$OUTPUT = $newContent;
		} // END - while

		// Output code here, DO NOT REMOVE! ;-)
		OUTPUT_RAW($OUTPUT);
	} elseif ((OUTPUT_MODE == "render") && (!empty($OUTPUT))) {
		// Rewrite links when rewrite extension is active
		if ((EXT_IS_ACTIVE("rewrite")) && (function_exists('REWRITE_LINKS')) && ($CSS != "1") && ($CSS != "-1")) {
			$OUTPUT = REWRITE_LINKS($OUTPUT);
		} // END - if

		// Compile and run finished rendered HTML code
		while (strpos($OUTPUT, '{!') > 0) {
			$eval = "\$OUTPUT = \"".COMPILE_CODE(addslashes($OUTPUT))."\";";
			eval($eval);
		} // END - while

		// Output code here, DO NOT REMOVE! ;-)
		OUTPUT_RAW($OUTPUT);
	}
}

// Output the raw HTML code
function OUTPUT_RAW ($HTML) {
	// Output stripped HTML code to avoid broken JavaScript code, etc.
	echo stripslashes(stripslashes($HTML));

	// Flush the output if only _OB_CACHING is not "on"
	if (_OB_CACHING != "on") {
		// Flush it
		flush();
	} // END - if
}

// Add a fatal error message to the queue array
function ADD_FATAL ($message, $extra="") {
	global $FATAL;

	if (empty($extra)) {
		// Regular text message to add to $FATAL
		$FATAL[] = $message;
	} else {
		// $message is text with a mask plus extras to insert into the text
		$message = sprintf($message, $extra);
		$FATAL[] = $message;
	}

	// Log fatal messages away
	DEBUG_LOG(__FUNCTION__, __LINE__, " message={$message}");
}

// Load a template file and return it's content (only it's name; do not use ' or ")
function LOAD_TEMPLATE($template, $return=false, $content=array()) {
	// Add more variables which you want to use in your template files
	global $DATA, $_CONFIG, $username;

	// Make all template names lowercase
	$template = strtolower($template);

	// Count the template load
	if (getConfig('num_templates') == null) {
		$_CONFIG['num_templates'] = 1;
	} else {
		$_CONFIG['num_templates']++;
	}

	// Prepare IP number and User Agent
	$REMOTE_ADDR     = GET_REMOTE_ADDR();
	if (!defined('REMOTE_ADDR')) define('REMOTE_ADDR', $REMOTE_ADDR);
	$HTTP_USER_AGENT = GET_USER_AGENT();

	// Init some data
	$ret = "";
	if (empty($GLOBALS['refid'])) $GLOBALS['refid'] = 0;

	// @DEPRECATED Try to rewrite the if() condition
	if ($template == "member_support_form") {
		// Support request of a member
		$result = SQL_QUERY_ESC("SELECT userid, gender, surname, family, email FROM `"._MYSQL_PREFIX."_user_data` WHERE userid=%s LIMIT 1",
			array($GLOBALS['userid']), __FILE__, __LINE__);

		// Is content an array?
		if (is_array($content)) {
			// Merge data
			$content = merge_array($content, SQL_FETCHARRAY($result));

			// Translate gender
			$content['gender'] = TRANSLATE_GENDER($content['gender']);
		} else {
			// DEPRECATED: Load data in direct variables
			list($gender, $surname, $family, $email) = SQL_FETCHROW($result);

			// Translate gender
			$gender = TRANSLATE_GENDER($gender);
			DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("DEPRECATION-WARNING: content is not array (%s).", gettype($content)));
		}

		// Free result
		SQL_FREERESULT($result);
	} // END - if

	// Generate date/time string
	$date_time = MAKE_DATETIME(time(), "1");

	// Base directory
	$BASE = sprintf("%stemplates/%s/html/", PATH, GET_LANGUAGE());
	$MODE = "";

	// Check for admin/guest/member templates
	if (strpos($template, "admin_") > -1) {
		// Admin template found
		$MODE = "admin/";
	} elseif (strpos($template, "guest_") > -1) {
		// Guest template found
		$MODE = "guest/";
	} elseif (strpos($template, "member_") > -1) {
		// Member template found
		$MODE = "member/";
	} elseif (strpos($template, "install_") > -1) {
		// Installation template found
		$MODE = "install/";
	} elseif (strpos($template, "ext_") > -1) {
		// Extension template found
		$MODE = "ext/";
	} elseif (strpos($template, "la_") > -1) {
		// "Logical-area" template found
		$MODE = "la/";
	} else {
		// Test for extension
		$test = substr($template, 0, strpos($template, "_"));
		if (EXT_IS_ACTIVE($test)) {
			// Set extra path to extension's name
			$MODE = $test."/";
		}
	}

	////////////////////////
	// Generate file name //
	////////////////////////
	$file = $BASE.$MODE.$template.".tpl";

	if ((!empty($GLOBALS['what'])) && ((strpos($template, "_header") > 0) || (strpos($template, "_footer") > 0)) && (($MODE == "guest/") || ($MODE == "member/") || ($MODE == "admin/"))) {
		// Select what depended header/footer template file for admin/guest/member area
		$file2 = sprintf("%s%s%s_%s.tpl",
			$BASE,
			$MODE,
			$template,
			SQL_ESCAPE($GLOBALS['what'])
		);

		// Probe for it...
		if (FILE_READABLE($file2)) $file = $file2;

		// Remove variable from memory
		unset($file2);
	}

	// Does the special template exists?
	if (!FILE_READABLE($file)) {
		// Reset to default template
		$file = $BASE.$template.".tpl";
	} // END - if

	// Now does the final template exists?
	if (FILE_READABLE($file)) {
		// The local file does exists so we load it. :)
		$tmpl_file = implode("", file($file));

		// Replace ' to our own chars to preventing them being quoted
		while (strpos($tmpl_file, "\'") !== false) { $tmpl_file = str_replace("\'", '{QUOT}', $tmpl_file); }

		// Do we have to compile the code?
		$ret = "";
		if ((strpos($tmpl_file, "\$") !== false) || (strpos($tmpl_file, '{--') !== false) || (strpos($tmpl_file, '--}') > 0)) {
			// Okay, compile it!
			$tmpl_file = "\$ret=\"".COMPILE_CODE(addslashes($tmpl_file))."\";";
			eval($tmpl_file);
		} else {
			// Simply return loaded code
			$ret = $tmpl_file;
		}

		// Add surrounding HTML comments to help finding bugs faster
		$ret = "<!-- Template ".$template." - Start -->\n".$ret."<!-- Template ".$template." - End -->\n";
	} elseif ((IS_ADMIN()) || ((isBooleanConstantAndTrue('mxchange_installing')) && (!isBooleanConstantAndTrue('mxchange_installed')))) {
		// Only admins shall see this warning or when installation mode is active
		$ret = "<br /><SPAN class=\"guest_failed\">".TEMPLATE_404."</SPAN><br />
(".basename($file).")<br />
<br />
".TEMPLATE_CONTENT."
<pre>".print_r($content, true)."</pre>
".TEMPLATE_DATA."
<pre>".print_r($DATA, true)."</pre>
<br /><br />";
	}

	// Remove content and data
	unset($content);
	unset($DATA);

	// Do we have some content to output or return?
	if (!empty($ret)) {
		// Not empty so let's put it out! ;)
		if ($return) {
			// Return the HTML code
			return $ret;
		} else {
			// Output direct
			OUTPUT_HTML($ret);
		}
	} elseif (isBooleanConstantAndTrue('DEBUG_MODE')) {
		// Warning, empty output!
		return "E:".$template."<br />\n";
	}
}

// Send mail out to an email address
function SEND_EMAIL($TO, $SUBJECT, $MSG, $HTML = "N", $FROM = "") {
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):TO={$TO},SUBJECT={$SUBJECT}<br />\n";

	// Compile subject line (for POINTS constant etc.)
	$eval = "\$SUBJECT = html_entity_decode(\"".COMPILE_CODE(addslashes($SUBJECT))."\");";
	eval($eval);

	// Set from header
	if ((!eregi("@", $TO)) && ($TO > 0)) {
		// Value detected, is the message extension installed?
		if (EXT_IS_ACTIVE("msg")) {
			ADD_MESSAGE_TO_BOX($TO, $SUBJECT, $MSG, $HTML);
			return;
		} else {
			// Load email address
			$result_email = SQL_QUERY_ESC("SELECT email FROM `"._MYSQL_PREFIX."_user_data` WHERE userid=%s LIMIT 1", array(bigintval($TO)), __FILE__, __LINE__);
			//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):numRows=".SQL_NUMROWS($result_email)."<br />\n";

			// Does the user exist?
			if (SQL_NUMROWS($result_email)) {
				// Load email address
				list($TO) = SQL_FETCHROW($result_email);
			} else {
				// Set webmaster
				$TO = WEBMASTER;
			}

			// Free result
			SQL_FREERESULT($result_email);
		}
	} elseif ("$TO" == "0") {
		// Is the webmaster!
		$TO = WEBMASTER;
	}
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):TO={$TO}<br />\n";

	// Check for PHPMailer or debug-mode
	if (!CHECK_PHPMAILER_USAGE()) {
		// Not in PHPMailer-Mode
		if (empty($FROM)) {
			// Load email header template
			$FROM = LOAD_EMAIL_TEMPLATE("header");
		} else {
			// Append header
			$FROM .= LOAD_EMAIL_TEMPLATE("header");
		}
	} elseif (isBooleanConstantAndTrue('DEBUG_MODE')) {
		if (empty($FROM)) {
			// Load email header template
			$FROM = LOAD_EMAIL_TEMPLATE("header");
		} else {
			// Append header
			$FROM .= LOAD_EMAIL_TEMPLATE("header");
		}
	}

	// Compile "TO"
	$eval = "\$TO = \"".COMPILE_CODE(addslashes($TO))."\";";
	eval($eval);

	// Compile "MSG"
	$eval = "\$MSG = \"".COMPILE_CODE(addslashes($MSG))."\";";
	eval($eval);

	// Fix HTML parameter (default is no!)
	if (empty($HTML)) $HTML = "N";
	if (isBooleanConstantAndTrue('DEBUG_MODE')) {
		// In debug mode we want to display the mail instead of sending it away so we can debug this part
		print("<pre>
".htmlentities(trim($FROM))."
To      : ".$TO."
Subject : ".$SUBJECT."
Message : ".$MSG."
</pre>\n");
	} elseif (($HTML == "Y") && (EXT_IS_ACTIVE("html_mail"))) {
		// Send mail as HTML away
		SEND_HTML_EMAIL($TO, $SUBJECT, $MSG, $FROM);
	} elseif (!empty($TO)) {
		// Send Mail away
		SEND_RAW_EMAIL($TO, $SUBJECT, $MSG, $FROM);
	} elseif ($HTML == "N") {
		// Problem found!
		SEND_RAW_EMAIL(WEBMASTER, "[PROBLEM:]".$SUBJECT, $MSG, $FROM);
	}
}

// Check if legacy or PHPMailer command
// @private
function CHECK_PHPMAILER_USAGE() {
	return ((defined('SMTP_HOSTNAME')) && (defined('SMTP_USER')) && (defined('SMTP_PASSWORD')) && (SMTP_HOSTNAME != "") && (SMTP_USER != ""));
}

/*
 * Send out a raw email with PHPMailer class or legacy mail() command
 */
function SEND_RAW_EMAIL ($to, $subject, $msg, $from) {
	// Shall we use PHPMailer class or legacy mode?
	if (CHECK_PHPMAILER_USAGE()) {
		// Use PHPMailer class with SMTP enabled
		require_once(PATH."inc/phpmailer/class.phpmailer.php");
		require_once(PATH."inc/phpmailer/class.smtp.php");

		// get new instance
		$mail = new PHPMailer();
		$mail->PluginDir  = sprintf("%sinc/phpmailer/", PATH);

		$mail->IsSMTP();
		$mail->SMTPAuth   = true;
		$mail->Host       = SMTP_HOSTNAME;
		$mail->Port       = 25;
		$mail->Username   = SMTP_USER;
		$mail->Password   = SMTP_PASSWORD;
		if (empty($from)) {
			$mail->From = WEBMASTER;
		} else {
			$mail->From = $from;
		}
		$mail->FromName   = MAIN_TITLE;
		$mail->Subject    = $subject;
		if ((EXT_IS_ACTIVE("html_mail")) && (strip_tags($msg) != $msg)) {
			$mail->Body       = $msg;
			$mail->AltBody    = "Your mail program required HTML support to read this mail!";
			$mail->WordWrap   = 70;
			$mail->IsHTML(true);
		} else {
			$mail->Body       = html_entity_decode($msg);
		}
		$mail->AddAddress($to, "");
		$mail->AddReplyTo(WEBMASTER,MAIN_TITLE);
		$mail->AddCustomHeader("Errors-To:".WEBMASTER);
		$mail->AddCustomHeader("X-Loop:".WEBMASTER);
		$mail->Send();
	} else {
		// Use legacy mail() command
		@mail($to, $subject, html_entity_decode($msg), $from);
	}
}
//

// Generate a password in a specified length or use default password length
function GEN_PASS($LEN = 0) {
	global $_CONFIG;
	if ($LEN == 0) $LEN = getConfig('pass_len');

	// Initialize array with all allowed chars
	$ABC = explode(",", "a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,0,1,2,3,4,5,6,7,8,9,-,+,_,/");

	// Initialize randomizer
	mt_srand((double) microtime() * 1000000);

	// Start creating password
	$PASS = "";
	for ($i = 0; $i < $LEN; $i++) {
		$PASS .= $ABC[mt_rand(0, sizeof($ABC) -1)];
	} // END - for

	// When the size is below 40 we can also add additional security by scrambling it
	if (strlen($PASS) <= 40) {
		// Also scramble the password
		$PASS = scrambleString($PASS);
	} // END - if

	// Return the password
	return $PASS;
}
//
function MAKE_DATETIME ($time, $mode="0")
{
	if ($time == 0) {
		// Never happend
		return NEVER_HAPPENED;
	} else {
		// Filter out numbers
		$time = bigintval($time);
	}

	switch (GET_LANGUAGE())
	{
	case "de": // German date / time format
		switch ($mode) {
			case "0": $ret = date("d.m.Y \u\m H:i \U\h\\r", $time); break;
			case "1": $ret = strtolower(date("d.m.Y - H:i", $time)); break;
			case "2": $ret = date("d.m.Y|H:i", $time); break;
			case "3": $ret = date("d.m.Y", $time); break;
			default:
				DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Invalid date mode %s detected.", $mode));
				break;
		}
		break;

	default:	// Default is the US date / time format!
		switch ($mode) {
			case "0": $ret = date("r", $time); break;
			case "1": $ret = date("Y-m-d - g:i A", $time); break;
			case "2": $ret = date("y-m-d|H:i", $time); break;
			case "3": $ret = date("y-m-d", $time); break;
			default:
				DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Invalid date mode %s detected.", $mode));
				break;
		}
	}
	return $ret;
}

// Translates the american decimal dot into a german comma
function TRANSLATE_COMMA ($dotted, $cut=true, $max=0) {
	global $_CONFIG;

	// Default is 3 you can change this in admin area "Misc -> Misc Options"
	if (getConfig('max_comma') == null) $_CONFIG['max_comma'] = "3";

	// Use from config is default
	$maxComma = getConfig('max_comma');

	// Use from parameter?
	if ($max > 0) $maxComma = $max;

	// Cut zeros off?
	if (($cut) && ($max == 0)) {
		// Test for commata if in cut-mode
		$com = explode(".", $dotted);
		if (count($com) < 2) {
			// Don't display commatas even if there are none... ;-)
			$maxComma = 0;
		}
	} // END - if

	// Debug log
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, "dotted={$dotted},maxComma={$maxComma}");

	// Translate it now
	switch (GET_LANGUAGE()) {
	case "de":
		$dotted = number_format($dotted, $maxComma, ",", ".");
		break;

	default:
		$dotted = number_format($dotted, $maxComma, ".", ",");
		break;
	}

	// Return translated value
	return $dotted;
}

//
function DEREFERER ($URL) {
	// Don't de-refer our own links!
	if (substr($URL, 0, strlen(URL)) != URL) {
		// De-refer this link
		$URL = URL."/modules.php?module=loader&amp;url=".urlencode(base64_encode(gzcompress($URL)));
	} // END - if

	// Return link
	return $URL;
}

//
function TRANSLATE_GENDER ($gender) {
	switch ($gender)
	{
		case "M": $ret = GENDER_M; break;
		case "F": $ret = GENDER_F; break;
		case "C": $ret = GENDER_C; break;
		default : $ret = $gender; break;
	}
	return $ret;
}
//
function FRAMETESTER($URL) {
	// Prepare frametester URL
	$frametesterUrl = sprintf("%s/modules.php?module=frametester&amp;url=%s",
		URL,
		urlencode(base64_encode(gzcompress(COMPILE_CODE($URL))))
	);
	return $frametesterUrl;
}
//
function SELECTION_COUNT($array) {
	$ret = 0;
	if (is_array($array)) {
		foreach ($array as $key => $sel) {
			if (!empty($sel)) $ret++;
		}
	}
	return $ret;
}
//
function IMG_CODE ($code, $type, $DATA, $uid) {
	return "<IMG border=\"0\" alt=\"Code\" src=\"".URL."/mailid_top.php?uid=".$uid."&amp;".$type."=".$DATA."&amp;mode=img&amp;code=".$code."\">";
}
//
function TRANSLATE_STATUS($status) {
	switch ($status)
	{
	case "UNCONFIRMED":
		$ret = ACCOUNT_UNCONFIRMED;
		break;

	case "CONFIRMED":
		$ret = ACCOUNT_CONFIRMED;
		break;

	case "LOCKED":
		$ret = ACCOUNT_LOCKED;
		break;

	case "":
	case null:
		$ret = ACCOUNT_DELETED;
		break;

	default:
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Unknown status %s detected.", $status));
		$ret = UNKNOWN_STATUS_1.$status.UNKNOWN_STATUS_2;
		break;
	}
	return $ret;
}
//
function GET_LANGUAGE() {
	global $cacheArray;

	// Set default return value to default language from config
	$ret = DEFAULT_LANG;

	// Init variable
	$lang = "";

	// Is the variable set
	if (!empty($_GET['mx_lang'])) {
		// Accept only first 2 chars
		$lang = substr($_GET['mx_lang'], 0, 2);
	} elseif (isset($cacheArray['language'])) {
		// Use cached
		$ret = $cacheArray['language'];
	} elseif (!empty($lang)) {
		// Check if main language file does exist
		if (FILE_READABLE(PATH."inc/language/".$lang.".php")) {
			// Okay found, so let's update cookies
			SET_LANGUAGE($lang);
		}
	} elseif (!isSessionVariableSet('mx_lang')) {
		// Return stored value from cookie
		$ret = get_session('mx_lang');

		// Fixes a warning before the session has the mx_lang constant
		if (empty($ret)) $ret = DEFAULT_LANG;
	}

	// Cache entry
	$cacheArray['language'] = $ret;

	// Return value
	return $ret;
}
//
function SET_LANGUAGE($lang) {
	global $_CONFIG;

	// Accept only first 2 chars!
	$lang = substr(SQL_ESCAPE(strip_tags($lang)), 0, 2);

	// Set cookie
	set_session("mx_lang", $lang);
}
//
function LOAD_EMAIL_TEMPLATE($template, $content=array(), $UID="0") {
	global $DATA, $_CONFIG, $REPLACER;

	// Make sure all template names are lowercase!
	$template = strtolower($template);

	// Default "nickname" if extension is not installed
	$nick = "---";

	// Keept for backward-compatiblity (please replace these variables against our new {!CONST!} syntax!)
	// No longer used: $MAIN_TITLE = MAIN_TITLE; $URL = URL; $WEBMASTER = WEBMASTER;

	// Prepare IP number and User Agent
	$REMOTE_ADDR     = GET_REMOTE_ADDR();
	$HTTP_USER_AGENT = GET_USER_AGENT();

	// Default admin
	$ADMIN = MAIN_TITLE;

	// Is the admin logged in?
	if (IS_ADMIN()) {
		// Get admin id
		$aid = GET_CURRENT_ADMIN_ID();

		// Load Admin data
		$ADMIN = GET_ADMIN_EMAIL($aid);
	} // END - if

	// Neutral email address is default
	$email = WEBMASTER;

	// Expiration in a nice output format
	if (getConfig('auto_purge') == 0) {
		// Will never expire!
		$EXPIRATION = MAIL_WILL_NEVER_EXPIRE;
	} elseif (function_exists('CREATE_FANCY_TIME')) {
		// Create nice date string
		$EXPIRATION = CREATE_FANCY_TIME(getConfig('auto_purge'));
	} else {
		// Display days only
		$EXPIRATION = round(getConfig('auto_purge')/60/60/24)." "._DAYS;
	}

	// Is content an array?
	if (is_array($content)) {
		// Add expiration to array, $EXPIRATION is now deprecated!
		$content['expiration'] = $EXPIRATION;
	} // END - if

	// Load user's data
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):UID={$UID},template={$template},content[]=".gettype($content)."<br />\n";
	if (($UID > 0) && (is_array($content))) {
		// If nickname extension is installed, fetch nickname as well
		if (EXT_IS_ACTIVE("nickname")) {
			//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):NICKNAME!<br />\n";
			// Load nickname
			$result = SQL_QUERY_ESC("SELECT surname, family, gender, email, nickname FROM `"._MYSQL_PREFIX."_user_data` WHERE userid=%s LIMIT 1",
				array(bigintval($UID)), __FILE__, __LINE__);
		} else {
			//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):NO-NICK!<br />\n";
			/// Load normal data
			$result = SQL_QUERY_ESC("SELECT surname, family, gender, email FROM `"._MYSQL_PREFIX."_user_data` WHERE userid=%s LIMIT 1",
				array(bigintval($UID)), __FILE__, __LINE__);
		}

		// Fetch and merge data
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):content()=".count($content)." - PRE<br />\n";
		$content = merge_array($content, SQL_FETCHARRAY($result));
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):content()=".count($content)." - AFTER<br />\n";

		// Free result
		SQL_FREERESULT($result);
	} // END - if

	// Translate M to male or F to female if present
	if (isset($content['gender'])) $content['gender'] = TRANSLATE_GENDER($content['gender']);

	// Overwrite email from data if present
	if (isset($content['email'])) $email = $content['email'];

	// Store email for some functions in global data array
	$DATA['email'] = $email;

	// Base directory
	$BASE = sprintf("%stemplates/%s/emails/", PATH, GET_LANGUAGE());

	// Check for admin/guest/member templates
	if (strpos($template, "admin_") > -1) {
		// Admin template found
		$file = $BASE."admin/".$template.".tpl";
	} elseif (strpos($template, "guest_") > -1) {
		// Guest template found
		$file = $BASE."guest/".$template.".tpl";
	} elseif (strpos($template, "member_") > -1) {
		// Member template found
		$file = $BASE."member/".$template.".tpl";
	} else {
		// Test for extension
		$test = substr($template, 0, strpos($template, "_"));
		if (EXT_IS_ACTIVE($test)) {
			// Set extra path to extension's name
			$file = $BASE.$test."/".$template.".tpl";
		} else {
			// No special filename
			$file = $BASE.$template.".tpl";
		}
	}

	// Does the special template exists?
	if (!FILE_READABLE($file)) {
		// Reset to default template
		$file = $BASE.$template.".tpl";
	} // END - if

	// Now does the final template exists?
	$newContent = "";
	if (FILE_READABLE($file)) {
		// The local file does exists so we load it. :)
		$tmpl_file = implode("", file($file));
		$tmpl_file = addslashes($tmpl_file);

		// Run code
		$tmpl_file = "\$newContent=html_entity_decode(\"".COMPILE_CODE($tmpl_file)."\");";
		@eval($tmpl_file);
	} elseif (!empty($template)) {
		// Template file not found!
		$newContent = TEMPLATE_404.": ".$template."<br />
".TEMPLATE_CONTENT."
<pre>".print_r($content, true)."</pre>
".TEMPLATE_DATA."
<pre>".print_r($DATA, true)."</pre>
<br /><br />";

		// Debug mode not active? Then remove the HTML tags
		if (!DEBUG_MODE) $newContent = strip_tags($newContent);
	} else {
		// No template name supplied!
		$newContent = NO_TEMPLATE_SUPPLIED;
	}

	// Is there some content?
	if (empty($newContent)) {
		// Compiling failed
		$newContent = "Compiler error for template {$template}!\nUncompiled content:\n".$tmpl_file;
		if (function_exists('error_get_last')) $newContent .= "\n--------------------------------------\nDebug:\n".print_r(error_get_last(), true)."--------------------------------------\nPlease don't alter these informations!\nThanx.";
	} // END - if

	// Remove content and data
	unset($content);
	unset($DATA);

	// Return compiled content
	return COMPILE_CODE($newContent);
}
//
function MAKE_TIME($H, $M, $S, $stamp) {
	// Extract day, month and year from given timestamp
	$DAY   = date("d", $stamp);
	$MONTH = date("m", $stamp);
	$YEAR  = date('Y', $stamp);

	// Create timestamp for wished time which depends on extracted date
	return mktime($H, $M, $S, $MONTH, $DAY, $YEAR);
}
//
function LOAD_URL($URL, $addUrlData=true) {
	global $CSS, $_CONFIG, $footer;

	// Compile out URI codes
	$URL = COMPILE_CODE($URL);

	// Check if http(s):// is there
	if ((substr($URL, 0, 7) != "http://") && (substr($URL, 0, 8) != "https://")) {
		// Make all URLs full-qualified
		$URL = URL."/".$URL;
	}

	// Get output buffer
	$OUTPUT = ob_get_contents();

	// Clear it only if there is content
	if (!empty($OUTPUT)) {
		ob_end_clean();
	} // END - if

	// Add some data to URL if cookies are not accepted
	if (((!defined('__COOKIES')) || (!__COOKIES)) && ($addUrlData)) $URL = ADD_URL_DATA($URL);

	// Probe for bot from search engine
	if ((eregi("spider", GET_USER_AGENT())) || (eregi("bot", GET_USER_AGENT())) || (eregi("spider", GET_USER_AGENT()))) {
		// Search engine bot detected so let's rewrite many chars for the link
		$URL = htmlentities(strip_tags($URL), ENT_QUOTES);

		// Output new location link as anchor
		OUTPUT_HTML("<A href=\"".$URL."\">".$URL."</A>");
	} elseif (!headers_sent()) {
		// Load URL when headers are not sent
		/*
		print("<pre>");
		debug_print_backtrace();
		die("</pre>URL={$URL}");
		*/
		@header ("Location: ".str_replace("&amp;", "&", $URL));
	} else {
		// Output error message
		include(PATH."inc/header.php");
		LOAD_TEMPLATE("redirect_url", false, str_replace("&amp;", "&", $URL));
		include(PATH."inc/footer.php");
	}
	exit();
}
//
function COMPILE_CODE($code, $simple = false, $constants = true, $full = true) {
	global $SEC_CHARS, $URL_CHARS;
	// Is the code a string?
	if (!is_string($code)) {
		// Silently return it
		return $code;
	} // END - if

	$ARRAY = $SEC_CHARS;

	// Select smaller set of chars to replace when we e.g. want to compile URLs
	if (!$full) $ARRAY = $URL_CHARS;

	// Compile constants
	if ($constants) {
		// BEFORE 0.2.1 : Language and data constants
		// WITH 0.2.1+  : Only language constants
		$code = str_replace('{--','".', str_replace('--}','."', $code));

		// BEFORE 0.2.1 : Not used
		// WITH 0.2.1+  : Data constants
		$code = str_replace('{!','".', str_replace("!}", '."', $code));
	} // END - if

	// Compile QUOT and other non-HTML codes
	foreach ($ARRAY['to'] as $k => $to) {
		// Do the reversed thing as in inc/libs/security_functions.php
		$code = str_replace($to, $ARRAY['from'][$k], $code);
	} // END - foreach

	// But shall I keep simple quotes for later use?
	if ($simple) $code = str_replace("\'", '{QUOT}', $code);

	// Find $content[bla][blub] entries
	@preg_match_all('/\$(content|DATA)((\[([a-zA-Z0-9-_]+)\])*)/', $code, $matches);

	// Are some matches found?
	if ((count($matches) > 0) && (count($matches[0]) > 0)) {
		// Replace all matches
		$matchesFound = array();
		foreach ($matches[0] as $key => $match) {
			// Fuzzy look has failed by default
			$fuzzyFound = false;

			// Fuzzy look on match if already found
			foreach ($matchesFound as $found => $set) {
				// Get test part
				$test = substr($found, 0, strlen($match));

				// Does this entry exist?
				//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):found={$found},match={$match},set={$set}<br />\n";
				if ($test == $match) {
					// Match found!
					//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):fuzzyFound!<br />\n";
					$fuzzyFound = true;
					break;
				} // END - if
			} // END - foreach

			// Skip this entry?
			if ($fuzzyFound) continue;

			// Take all string elements
			if ((is_string($matches[4][$key])) && (!isset($matchesFound[$match])) && (!isset($matchesFound[$key."_".$matches[4][$key]]))) {
				// Replace it in the code
				//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):key={$key},match={$match}<br />\n";
				$newMatch = str_replace("[".$matches[4][$key]."]", "['".$matches[4][$key]."']", $match);
				$code = str_replace($match, "\".".$newMatch.".\"", $code);
				$matchesFound[$key."_".$matches[4][$key]] = 1;
				$matchesFound[$match] = 1;
			} elseif (!isset($matchesFound[$match])) {
				// Not yet replaced!
				//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):match={$match}<br />\n";
				$code = str_replace($match, "\".".$match.".\"", $code);
				$matchesFound[$match] = 1;
			}
		} // END - foreach
	} // END - if

	// Return compiled code
	return $code;
}
//
/************************************************************************
 *                                                                      *
 * Gaenderter Sortier-Algorythmus, $array wird nach dem Array (!)       *
 * $a_sort sortiert:                                                    *
 *                                                                      *
 * $array - Das 3-dimensionale Array, das paralell sortiert werden soll *
 * $a_sort - Array, das die Sortiereihenfolge der ersten Elementeben    *
 * $primary_key - Prim.rschl.ssel aus $a_sort, nach dem sortiert wird   *
 * $order - Sortiereihenfolge: -1 = A-Z, 0 = keine, 1 = Z-A             *
 * $nums - true = Als Zahlen sortieren, false = Als Zeichen sortieren   *
 *                                                                      *
 * $a_sort muss Elemente enthalten, deren Wert Schluessel von $array    *
 * sind... Klingt kompliziert, suchen Sie mal mein Beispiel, dann sehen *
 * Sie, dass es doch nicht so schwer ist! :-)                           *
 *                                                                      *
 ************************************************************************/
function array_pk_sort(&$array, $a_sort, $primary_key = 0, $order = -1, $nums = false)
{
	$dummy = $array;
	while ($primary_key < count($a_sort)) {
		foreach ($dummy[$a_sort[$primary_key]] as $key => $value) {
			foreach ($dummy[$a_sort[$primary_key]] as $key2 => $value2) {
				$match = false;
				if (!$nums) {
					// Sort byte-by-byte (also numbers will be interpreted as chars! E.g.: "9" > "10")
					if (($key != $key2) && (strcmp(strtolower($dummy[$a_sort[$primary_key]][$key]), strtolower($dummy[$a_sort[$primary_key]][$key2])) == $order)) $match = true;
				} elseif ($key != $key2) {
					// Sort numbers (E.g.: 9 < 10)
					if (($dummy[$a_sort[$primary_key]][$key] < $dummy[$a_sort[$primary_key]][$key2]) && ($order == -1)) $match = true;
					if (($dummy[$a_sort[$primary_key]][$key] > $dummy[$a_sort[$primary_key]][$key2]) && ($order == 1))  $match = true;
				}

				if ($match) {
					// We have found two different values, so let's sort whole array
					foreach ($dummy as $sort_key => $sort_val) {
						$t                       = $dummy[$sort_key][$key];
						$dummy[$sort_key][$key]  = $dummy[$sort_key][$key2];
						$dummy[$sort_key][$key2] = $t;
						unset($t);
					} // END - foreach
				} // END - if
			} // END - foreach
		} // END - foreach

		// Count one up
		$primary_key++;
	} // END - while

	// Write back sorted array
	$array = $dummy;
}
//
function ADD_SELECTION($type, $DEFAULT, $prefix="", $id="0") {
	global $MONTH_DESCR, $_CONFIG;
	$OUT = "";

	if ($type == "yn") {
		// This is a yes/no selection only!
		if ($id > 0) $prefix .= "[".$id."]";
		$OUT .= "    <SELECT name=\"".$prefix."\" class=\"register_select\" size=\"1\">\n";
	} else {
		// Begin with regular selection box here
		if (!empty($prefix)) $prefix .= "_";
		$type2 = $type;
		if ($id > 0) $type2 .= "[".$id."]";
		$OUT .= "    <SELECT name=\"".strtolower($prefix.$type2)."\" class=\"register_select\" size=\"1\">\n";
	}

	switch ($type) {
	case "day": // Day
		for ($idx = 1; $idx < 32; $idx++) {
			$OUT .= "<OPTION value=\"".$idx."\"";
			if ($DEFAULT == $idx) $OUT .= " selected=\"selected\"";
			$OUT .= ">".$idx."</OPTION>\n";
		} // END - for
		break;

	case "month": // Month
		foreach ($MONTH_DESCR as $month => $descr) {
			$OUT .= "<OPTION value=\"".$month."\"";
			if ($DEFAULT == $month) $OUT .= " selected=\"selected\"";
			$OUT .= ">".$descr."</OPTION>\n";
		} // END - for
		break;

	case "year": // Year
		// Get current year
		$YEAR = date('Y', time());

		// Use configured min age or fixed?
		if (GET_EXT_VERSION("other") >= "0.2.1") {
			// Configured
			$startYear = $YEAR - getConfig('min_age');
		} else {
			// Fixed 16 years
			$startYear = $YEAR - 16;
		}

		// Calculate earliest year (100 years old people can still enter Internet???)
		$minYear = $YEAR - 100;

		// Check if the default value is larger than minimum and bigger than actual year
		if (($DEFAULT > $minYear) && ($DEFAULT >= $YEAR)) {
			for ($idx = $YEAR; $idx < ($YEAR + 11); $idx++) {
				$OUT .= "<OPTION value=\"".$idx."\"";
				if ($DEFAULT == $idx) $OUT .= " selected=\"selected\"";
				$OUT .= ">".$idx."</OPTION>\n";
			} // END - for
		} elseif ($DEFAULT == -1) {
			// Current year minus 1
			for ($idx = $startYear; $idx <= ($YEAR + 1); $idx++)
			{
				$OUT .= "<OPTION value=\"".$idx."\">".$idx."</OPTION>\n";
			}
		} else {
			// Get current year and subtract the configured minimum age
			$OUT .= "<OPTION value=\"".($minYear - 1)."\">&lt;".$minYear."</OPTION>\n";
			// Calculate earliest year depending on extension version
			if (GET_EXT_VERSION("other") >= "0.2.1") {
				// Use configured minimum age
				$YEAR = date('Y', time()) - getConfig('min_age');
			} else {
				// Use fixed 16 years age
				$YEAR = date('Y', time()) - 16;
			}

			// Construct year selection list
			for ($idx = $minYear; $idx <= $YEAR; $idx++) {
				$OUT .= "<OPTION value=\"".$idx."\"";
				if ($DEFAULT == $idx) $OUT .= " selected=\"selected\"";
				$OUT .= ">".$idx."</OPTION>\n";
			} // END - for
		}
		break;

	case "sec":
	case "min":
		for ($idx = 0; $idx < 60; $idx+=5) {
			if (strlen($idx) == 1) $idx = "0".$idx;
			$OUT .= "<OPTION value=\"".$idx."\"";
			if ($DEFAULT == $idx) $OUT .= " selected=\"selected\"";
			$OUT .= ">".$idx."</OPTION>\n";
		} // END - for
		break;

	case "hour":
		for ($idx = 0; $idx < 24; $idx++) {
			if (strlen($idx) == 1) $idx = "0".$idx;
			$OUT .= "<OPTION value=\"".$idx."\"";
			if ($DEFAULT == $idx) $OUT .= " selected=\"selected\"";
			$OUT .= ">".$idx."</OPTION>\n";
		} // END - for
		break;

	case "yn":
		$OUT .= "<OPTION value=\"Y\"";
		if ($DEFAULT == "Y") $OUT .= " selected=\"selected\"";
		$OUT .= ">".YES."</OPTION>\n<OPTION value=\"N\"";
		if ($DEFAULT == "N") $OUT .= " selected=\"selected\"";
		$OUT .= ">".NO."</OPTION>\n";
		break;
	}
	$OUT .= "    </SELECT>\n";
	return $OUT;
}
//
function TRANSLATE_YESNO($yn)
{
	switch ($yn)
	{
		case "Y": $yn = YES; break;
		case "N": $yn = NO; break;
		default : $yn = "??? (".$yn.")"; break;
	}
	return $yn;
}
//
// Deprecated : $length
// Optional   : $DATA
//
function GEN_RANDOM_CODE($length, $code, $uid, $DATA="") {
	global $_CONFIG;

	// Fix missing _MAX constant
	if (!defined('_MAX')) define('_MAX', 15235);

	// Build server string
	$server = $_SERVER['PHP_SELF'].":".GET_USER_AGENT().":".getenv('SERVER_SOFTWARE').":".GET_REMOTE_ADDR().":".":".filemtime(PATH."inc/databases.php");

	// Build key string
	$keys   = SITE_KEY.":".DATE_KEY;
	if (getConfig('secret_key') != null)  $keys .= ":".getConfig('secret_key');
	if (getConfig('file_hash') != null)   $keys .= ":".getConfig('file_hash');
	$keys .= ":".date("d-m-Y (l-F-T)", bigintval(getConfig('patch_ctime')));
	if (getConfig('master_salt') != null) $keys .= ":".getConfig('master_salt');

	// Build string from misc data
	$data   = $code.":".$uid.":".$DATA;

	// Add more additional data
	if (isSessionVariableSet('u_hash'))			$data .= ":".get_session('u_hash');
	if (isset($GLOBALS['userid']))				$data .= ":".$GLOBALS['userid'];
	if (isSessionVariableSet('lifetime'))		$data .= ":".get_session('lifetime');
	if (isSessionVariableSet('mxchange_theme'))	$data .= ":".get_session('mxchange_theme');
	if (isSessionVariableSet('mx_lang'))		$data .= ":".GET_LANGUAGE();
	if (isset($GLOBALS['refid']))				$data .= ":".$GLOBALS['refid'];

	// Calculate number for generating the code
	$a = $code + _ADD - 1;

	if (getConfig('master_hash') != null) {
		// Generate hash with master salt from modula of number with the prime number and other data
		$saltedHash = generateHash(($a % _PRIME).":".$server.":".$keys.":".$data.":".date("d-m-Y (l-F-T)", time()).":".$a, getConfig('master_salt'));

		// Create number from hash
		$rcode = hexdec(substr($saltedHash, strlen(getConfig('master_salt')), 9)) / abs(_MAX - $a + sqrt(_ADD)) / pi();
	} else {
		// Generate hash with "hash of site key" from modula of number with the prime number and other data
		$saltedHash = generateHash(($a % _PRIME).":".$server.":".$keys.":".$data.":".date("d-m-Y (l-F-T)", time()).":".$a, substr(sha1(SITE_KEY), 0, 8));

		// Create number from hash
		$rcode = hexdec(substr($saltedHash, 8, 9)) / abs(_MAX - $a + sqrt(_ADD)) / pi();
	}

	// At least 10 numbers shall be secure enought!
	$len = getConfig('code_length');
	if ($len == 0) $len = $length;
	if ($len == 0) $len = 10;

	// Cut off requested counts of number
	$return = substr(str_replace('.', "", $rcode), 0, $len);

	// Done building code
	return $return;
}
// Does only allow numbers
function bigintval($num, $castValue = true) {
	// Filter all numbers out
	$ret = preg_replace("/[^0123456789]/", "", $num);

	// Shall we cast?
	if ($castValue) $ret = (double)$ret;

	// Has the whole value changed?
	if ("".$ret."" != "".$num."") {
		// Log the values
		print("<pre>");
		debug_print_backtrace();
		die("</pre>");
		DEBUG_LOG(__FUNCTION__, __LINE__, " num={$num},ret={$ret}");
	} // END - if

	// Return result
	return $ret;
}
// Insert the code in $img_code into jpeg or PNG image
function GENERATE_IMAGE($img_code, $header=true) {
	global $_CONFIG;

	if ((strlen($img_code) > 6) || (empty($img_code)) || (getConfig('code_length') == 0)) {
		// Stop execution of function here because of over-sized code length
		return;
	} elseif (!$header) {
		// Return in an HTML code code
		return "<IMG src=\"".URL."/img.php?code=".$img_code."\">\n";
	}

	// Load image
	$img = sprintf("%s/theme/%s/images/code_bg.%s", PATH, GET_CURR_THEME(), getConfig('img_type'));
	if (FILE_READABLE($img)) {
		// Switch image type
		switch (getConfig('img_type'))
		{
		case "jpg":
			// Okay, load image and hide all errors
			$image = @imagecreatefromjpeg($img);
			break;

		case "png":
			// Okay, load image and hide all errors
			$image = @imagecreatefrompng($img);
			break;
		}
	} else {
		// Exit function here
		return;
	}

	// Generate text color (red/green/blue; 0 = dark, 255 = bright)
	$text_color = imagecolorallocate($image, 0, 0, 0);

	// Insert code into image
	imagestring($image, 5, 14, 2, $img_code, $text_color);

	// Return to browser
	header ("Content-Type: image/".getConfig('img_type'));

	// Output image with matching image factory
	switch (getConfig('img_type')) {
		case "jpg": imagejpeg($image); break;
		case "png": imagepng($image);  break;
	}

	// Remove image from memory
	imagedestroy($image);
}
// Create selection box or array of splitted timestamp
function CREATE_TIME_SELECTIONS ($timestamp, $prefix="", $display="", $align="center", $return_array=false) {
	global $_CONFIG;

	// Calculate 2-seconds timestamp
	$stamp = round($timestamp);
	//* DEBUG: */ print("*".$stamp."/".$timestamp."*<br />");

	// Do we have a leap year?
	$SWITCH = 0;
	$TEST = date('Y', time()) / 4;
	$M1 = date("m", time());
	$M2 = date("m", (time() + $timestamp));

	// If so and if current time is before 02/29 and estimated time is after 02/29 then add 86400 seconds (one day)
	if ((floor($TEST) == $TEST) && ($M1 == "02") && ($M2 > "02"))  $SWITCH = getConfig('one_day');

	// First of all years...
	$Y = abs(floor($timestamp / (31536000 + $SWITCH)));
	//* DEBUG: */ print("Y={$Y}<br />\n");
	// Next months...
	$M = abs(floor($timestamp / 2628000 - $Y * 12));
	//* DEBUG: */ print("M={$M}<br />\n");
	// Next weeks
	$W = abs(floor($timestamp / 604800 - $Y * ((365 + $SWITCH / getConfig('one_day')) / 7) - ($M / 12 * (365 + $SWITCH / getConfig('one_day')) / 7)));
	//* DEBUG: */ print("W={$W}<br />\n");
	// Next days...
	$D = abs(floor($timestamp / 86400 - $Y * (365 + $SWITCH / getConfig('one_day')) - ($M / 12 * (365 + $SWITCH / getConfig('one_day'))) - $W * 7));
	//* DEBUG: */ print("D={$D}<br />\n");
	// Next hours...
	$h = abs(floor($timestamp / 3600 - $Y * (365 + $SWITCH / getConfig('one_day')) * 24 - ($M / 12 * (365 + $SWITCH / getConfig('one_day')) * 24) - $W * 7 * 24 - $D * 24));
	//* DEBUG: */ print("h={$h}<br />\n");
	// Next minutes..
	$m = abs(floor($timestamp / 60 - $Y * (365 + $SWITCH / getConfig('one_day')) * 24 * 60 - ($M / 12 * (365 + $SWITCH / getConfig('one_day')) * 24 * 60) - $W * 7 * 24 * 60 - $D * 24 * 60 - $h * 60));
	//* DEBUG: */ print("m={$m}<br />\n");
	// And at last seconds...
	$s = abs(floor($timestamp - $Y * (365 + $SWITCH / getConfig('one_day')) * 24 * 3600 - ($M / 12 * (365 + $SWITCH / getConfig('one_day')) * 24 * 3600) - $W * 7 * 24 * 3600 - $D * 24 * 3600 - $h * 3600 - $m * 60));
	//* DEBUG: */ print("s={$s}<br />\n");

	// Is seconds zero and time is < 60 seconds?
	if (($s == 0) && ($timestamp < 60)) {
		// Fix seconds
		$s = round($timestamp);
	} // END - if

	//
	// Now we convert them in seconds...
	//
	if ($return_array) {
		// Just put all data in an array for later use
		$OUT = array(
			'YEARS'   => $Y,
			'MONTHS'  => $M,
			'WEEKS'   => $W,
			'DAYS'    => $D,
			'HOURS'   => $h,
			'MINUTES' => $m,
			'SECONDS' => $s
		);
	} else {
		// Generate table
		$OUT  = "<DIV align=\"".$align."\">\n";
		$OUT .= "<TABLE border=\"0\" cellspacing=\"0\" cellpadding=\"0\" class=\"admin_table dashed\">\n";
		$OUT .= "<TR>\n";

		if (ereg('Y', $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._YEARS."</STRONG></TD>\n";
		}

		if (ereg("M", $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._MONTHS."</STRONG></TD>\n";
		}

		if (ereg("W", $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._WEEKS."</STRONG></TD>\n";
		}

		if (ereg("D", $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._DAYS."</STRONG></TD>\n";
		}

		if (ereg("h", $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._HOURS."</STRONG></TD>\n";
		}

		if (ereg("m", $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._MINUTES."</STRONG></TD>\n";
		}

		if (ereg("s", $display) || (empty($display))) {
			$OUT .= "  <TD align=\"center\" class=\"admin_title bottom\"><STRONG class=\"tiny\">"._SECONDS."</STRONG></TD>\n";
		}

		$OUT .= "</TR>\n";
		$OUT .= "<TR>\n";

		if (ereg('Y', $display) || (empty($display))) {
			// Generate year selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_ye\" size=\"1\">\n";
			for ($idx = 0; $idx <= 10; $idx++) {
				$OUT .= "    <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $Y) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_ye\" value=\"0\">\n";
		}

		if (ereg("M", $display) || (empty($display))) {
			// Generate month selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_mo\" size=\"1\">\n";
			for ($idx = 0; $idx <= 11; $idx++)
			{
					$OUT .= "  <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $M) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_mo\" value=\"0\">\n";
		}

		if (ereg("W", $display) || (empty($display))) {
			// Generate week selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_we\" size=\"1\">\n";
			for ($idx = 0; $idx <= 4; $idx++) {
				$OUT .= "  <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $W) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_we\" value=\"0\">\n";
		}

		if (ereg("D", $display) || (empty($display))) {
			// Generate day selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_da\" size=\"1\">\n";
			for ($idx = 0; $idx <= 31; $idx++) {
				$OUT .= "  <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $D) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_da\" value=\"0\">\n";
		}

		if (ereg("h", $display) || (empty($display))) {
			// Generate hour selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_ho\" size=\"1\">\n";
			for ($idx = 0; $idx <= 23; $idx++) 	{
				$OUT .= "  <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $h) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_ho\" value=\"0\">\n";
		}

		if (ereg("m", $display) || (empty($display))) {
			// Generate minute selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_mi\" size=\"1\">\n";
			for ($idx = 0; $idx <= 59; $idx++) {
				$OUT .= "  <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $m) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_mi\" value=\"0\">\n";
		}

		if (ereg("s", $display) || (empty($display))) {
			// Generate second selection
			$OUT .= "  <TD align=\"center\"><SELECT class=\"mini_select\" name=\"".$prefix."_se\" size=\"1\">\n";
			for ($idx = 0; $idx <= 59; $idx++) {
				$OUT .= "  <OPTION class=\"mini_select\" value=\"".$idx."\"";
				if ($idx == $s) $OUT .= " selected default";
				$OUT .= ">".$idx."</OPTION>\n";
			}
			$OUT .= "  </SELECT></TD>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"".$prefix."_se\" value=\"0\">\n";
		}
		$OUT .= "</TR>\n";
		$OUT .= "</TABLE>\n";
		$OUT .= "</DIV>\n";
		// Return generated HTML code
	}
	return $OUT;
}
//
function CREATE_TIMESTAMP_FROM_SELECTIONS($prefix, $POST) {
	global $_CONFIG;
	$ret = 0;

	// Do we have a leap year?
	$SWITCH = 0;
	$TEST = date('Y', time()) / 4;
	$M1   = date("m", time());
	// If so and if current time is before 02/29 and estimated time is after 02/29 then add 86400 seconds (one day)
	if ((floor($TEST) == $TEST) && ($M1 == "02") && ($POST[$prefix."_mo"] > "02"))  $SWITCH = getConfig('one_day');
	// First add years...
	$ret += $POST[$prefix."_ye"] * (31536000 + $SWITCH);
	// Next months...
	$ret += $POST[$prefix."_mo"] * 2628000;
	// Next weeks
	$ret += $POST[$prefix."_we"] * 604800;
	// Next days...
	$ret += $POST[$prefix."_da"] * 86400;
	// Next hours...
	$ret += $POST[$prefix."_ho"] * 3600;
	// Next minutes..
	$ret += $POST[$prefix."_mi"] * 60;
	// And at last seconds...
	$ret += $POST[$prefix."_se"];
	// Return calculated value
	return $ret;
}
// Sends out mail to all administrators
// IMPORTANT: Please use SEND_ADMIN_NOTIFCATION() for now!
function SEND_ADMIN_EMAILS_PRO($subj, $template, $content, $UID) {
	// Trim template name
	$template = trim($template);

	// Load email template
	$msg = LOAD_EMAIL_TEMPLATE($template, $content, $UID);

	if (EXT_VERSION_IS_OLDER("admins", "0.4.0")) {
		// Older version detected!
		return SEND_ADMIN_EMAILS($subj, $msg);
	} // END - if

	// Check which admin shall receive this mail
	$result = SQL_QUERY_ESC("SELECT DISTINCT admin_id FROM "._MYSQL_PREFIX."_admins_mails WHERE mail_template='%s' ORDER BY admin_id",
		array($template), __FILE__, __LINE__);
	if (SQL_NUMROWS($result) == 0) {
		// Create new entry (to all admins)
		SQL_QUERY_ESC("INSERT INTO "._MYSQL_PREFIX."_admins_mails (admin_id, mail_template) VALUES (0, '%s')",
			array($template), __FILE__, __LINE__);
	} else {
		// Load admin IDs...
		$aids = array();
		while(list($aid) = SQL_FETCHROW($result)) {
			$aids[] = $aid;
		}

		// Free memory
		SQL_FREERESULT($result);

		// "implode" IDs and query string
		$aid = implode(",", $aids);
		if ($aid == "-1") {
			// Add line to userlog
			USERLOG_ADD_LINE($subj, $msg, $UID);
			return;
		} elseif ($aid == "0") {
			// Select all email adresses
			$result = SQL_QUERY("SELECT email FROM "._MYSQL_PREFIX."_admins ORDER BY id", __FILE__, __LINE__);
		} else {
			// If Admin-ID is not "to-all" select
			$result = SQL_QUERY_ESC("SELECT email FROM "._MYSQL_PREFIX."_admins WHERE id IN (%s) ORDER BY id", array($aid), __FILE__, __LINE__);
		}
	}

	// Load email addresses and send away
	while (list($email) = SQL_FETCHROW($result)) {
		SEND_EMAIL($email, $subj, $msg);
	}

	// Free memory
	SQL_FREERESULT($result);
}
//
function CREATE_FANCY_TIME ($stamp) {
	// Get data array with years/months/weeks/days/...
	$data = CREATE_TIME_SELECTIONS($stamp, "", "", "", true);
	$ret = "";
	foreach($data as $k => $v) {
		if ($v > 0) {
			// Value is greater than 0 "eval" data to return string
			$eval = "\$ret .= \", \".\$v.\" \"._".strtoupper($k).";";
			eval($eval);
			break;
		} // END - if
	} // END - foreach

	// Do we have something there?
	if (strlen($ret) > 0) {
		// Remove leading commata and space
		$ret = substr($ret, 2);
	} else {
		// Zero seconds
		$ret = "0 "._SECONDS;
	}

	// Return fancy time string
	return $ret;
}
//
function ADD_EMAIL_NAV($PAGES, $offset, $show_form, $colspan, $return=false) {
	$SEP = ""; $TOP = "";
	if (!$show_form) {
		$TOP = " top2";
		$SEP = "<TR><TD colspan=\"".$colspan."\" class=\"seperator\">&nbsp;</TD></TR>";
	}

	$NAV = "";
	for ($page = 1; $page <= $PAGES; $page++) {
		// Is the page currently selected or shall we generate a link to it?
		if (($page == $_GET['page']) || ((empty($_GET['page'])) && ($page == "1"))) {
			// Is currently selected, so only highlight it
			$NAV .= "<STRONG>-";
		} else {
			// Open anchor tag and add base URL
			$NAV .= "<A href=\"".URL."/modules.php?module=admin&amp;what=".$GLOBALS['what']."&amp;page=".$page."&amp;offset=".$offset;

			// Add userid when we shall show all mails from a single member
			if ((isset($_GET['u_id'])) && (bigintval($_GET['u_id']) > 0)) $NAV .= "&amp;u_id=".bigintval($_GET['u_id']);

			// Close open anchor tag
			$NAV .= "\">";
		}
		$NAV .= $page;
		if (($page == $_GET['page']) || ((empty($_GET['page'])) && ($page == "1"))) {
			// Is currently selected, so only highlight it
			$NAV .= "-</STRONG>";
		} else {
			// Close anchor tag
			$NAV .= "</A>";
		}

		// Add seperator if we have not yet reached total pages
		if ($page < $PAGES) $NAV .= "&nbsp;|&nbsp;";
	}

	// Define constants only once
	if (!defined('__NAV_OUTPUT')) {
		define('__NAV_OUTPUT' , $NAV);
		define('__NAV_COLSPAN', $colspan);
		define('__NAV_TOP'    , $TOP);
		define('__NAV_SEP'    , $SEP);
	}

	// Load navigation template
	$OUT = LOAD_TEMPLATE("admin_email_nav_row", true);

	if ($return) {
		// Return generated HTML-Code
		return $OUT;
	} else {
		// Output HTML-Code
		OUTPUT_HTML($OUT);
	}
}

// Extract host from script name
function EXTRACT_HOST (&$script) {
	// Use default SERVER_URL by default... ;) So?
	$url = SERVER_URL;

	// Is this URL valid?
	if (substr($script, 0, 7) == "http://") {
		// Use the hostname from script URL as new hostname
		$url = substr($script, 7);
		$extract = explode("/", $url);
		$url = $extract[0];
		// Done extracting the URL :)
	} // END - if

	// Extract host name
	$host = str_replace("http://", "", $url);
	if (ereg("/", $host)) $host = substr($host, 0, strpos($host, "/"));

	// Generate relative URL
	//* DEBUG: */ print("SCRIPT=".$script."<br />\n");
	if (substr(strtolower($script), 0, 7) == "http://") {
		// But only if http:// is in front!
		$script = substr($script, (strlen($url) + 7));
	} elseif (substr(strtolower($script), 0, 8) == "https://") {
		// Does this work?!
		$script = substr($script, (strlen($url) + 8));
	}

	//* DEBUG: */ print("SCRIPT=".$script."<br />\n");
	if (substr($script, 0, 1) == "/") $script = substr($script, 1);

	// Return host name
	return $host;
}

// Send a GET request
function GET_URL ($script) {
	// Compile the script name
	$script = COMPILE_CODE($script);

	// Extract host name from script
	$host = EXTRACT_HOST($script);

	// Generate GET request header
	$request  = "GET /" . trim($script) . " HTTP/1.1\r\n";
	$request .= "Host: " . $host . "\r\n";
	$request .= "Referer: " . URL . "/admin.php\r\n";
	$request .= "User-Agent: " . TITLE . "/" . FULL_VERSION . "\r\n";
	$request .= "Content-Type: text/plain\r\n";
	$request .= "Cache-Control: no-cache\r\n";
	$request .= "Connection: Close\r\n\r\n";

	// Send the raw request
	$response = SEND_RAW_REQUEST($host, $request);

	// Return the result to the caller function
	return $response;
}

// Send a POST request
function POST_URL ($script, $postData) {
	// Is postData an array?
	if (!is_array($postData)) {
		// Abort here
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("postData is not an array. Type: %s", gettype($postData)));
		return array("", "", "");
	} // END - if

	// Compile the script name
	$script = COMPILE_CODE($script);

	// Extract host name from script
	$host = EXTRACT_HOST($script);

	// Construct request
	$data = http_build_query($postData, '','&');

	// Generate POST request header
	$request  = "POST /" . trim($script) . " HTTP/1.1\r\n";
	$request .= "Host: " . $host . "\r\n";
	$request .= "Referer: " . URL . "/admin.php\r\n";
	$request .= "User-Agent: " . TITLE . "/" . FULL_VERSION . "\r\n";
	$request .= "Content-type: application/x-www-form-urlencoded\r\n";
	$request .= "Content-length: " . strlen($data) . "\r\n";
	$request .= "Cache-Control: no-cache\r\n";
	$request .= "Connection: Close\r\n\r\n";
	$request .= $data;

	// Send the raw request
	$response = SEND_RAW_REQUEST($host, $request);

	// Return the result to the caller function
	return $response;
}

// Sends a raw request to another host
function SEND_RAW_REQUEST ($host, $request) {
	global $_CONFIG;

	// Initialize array
	$response = array("", "", "");

	// Default is not to use proxy
	$useProxy = false;

	// Are proxy settins set?
	if ((getConfig('proxy_host') != "") && (getConfig('proxy_port') > 0)) {
		// Then use it
		$useProxy = true;
	} // END - if

	// Open connection
	//* DEBUG: */ die("SCRIPT=".$script."<br />\n");
	if ($useProxy) {
		$fp = @fsockopen(COMPILE_CODE(getConfig('proxy_host')), getConfig('proxy_port'), $errno, $errdesc, 30);
	} else {
		$fp = @fsockopen($host, 80, $errno, $errdesc, 30);
	}

	// Is there a link?
	if (!is_resource($fp)) {
		// Failed!
		return $response;
	} // END - if

	// Do we use proxy?
	if ($useProxy) {
		// Generate CONNECT request header
		$proxyTunnel  = "CONNECT ".$host.":80 HTTP/1.1\r\n";
		$proxyTunnel .= "Host: ".$host."\r\n";

		// Use login data to proxy? (username at least!)
		if (getConfig('proxy_username') != "") {
			// Add it as well
			$encodedAuth = base64_encode(COMPILE_CODE(getConfig('proxy_username')).":".COMPILE_CODE(getConfig('proxy_password')));
			$proxyTunnel .= "Proxy-Authorization: Basic ".$encodedAuth."\r\n";
		} // END - if

		// Add last new-line
		$proxyTunnel .= "\r\n";
		//* DEBUG: */ print("<strong>proxyTunnel=</strong><pre>".$proxyTunnel."</pre>");

		// Write request
		fputs($fp, $proxyTunnel);

		// Got response?
		if (feof($fp)) {
			// No response received
			return $response;
		} // END - if

		// Read the first line
		$resp = trim(fgets($fp, 10240));
		$respArray = explode(" ", $resp);
		if ((strtolower($respArray[0]) !== "http/1.0") || ($respArray[1] != "200")) {
			// Invalid response!
			return $response;
		} // END - if
	} // END - if

	// Write request
	fputs($fp, $request);

	// Read response
	while(!feof($fp)) {
		$response[] = trim(fgets($fp, 1024));
	} // END - while

	// Close socket
	fclose($fp);

	// Skip first empty lines
	$resp = $response;
	foreach ($resp as $idx => $line) {
		// Trim space away
		$line = trim($line);

		// Is this line empty?
		if (empty($line)) {
			// Then remove it
			array_shift($response);
		} else {
			// Abort on first non-empty line
			break;
		}
	} // END - foreach

	//* DEBUG: */ print("<strong>Response:</strong><pre>".print_r($response, true)."</pre>");

	// Proxy agent found?
	if ((substr(strtolower($response[0]), 0, 11) == "proxy-agent") && ($useProxy)) {
		// Proxy header detected, so remove two lines
		array_shift($response);
		array_shift($response);
	} // END - if

	// Was the request successfull?
	if ((!eregi("200 OK", $response[0])) || (empty($response[0]))) {
		// Not found / access forbidden
		$response = array("", "", "");
	} // END - if

	// Return response
	return $response;
}
// Taken from www.php.net eregi() user comments
function VALIDATE_EMAIL($email) {
	// Compile email
	$email = COMPILE_CODE($email);

	// Check first part of email address
	$first = "[-a-z0-9!#$%&\'*+/=?^_<{|}~]+(\.[-a-zA-Z0-9!#$%&\'*+/=?^_<{|}~]+)*";

	//  Check domain
	$domain = "[a-z0-9-]+(\.[a-z0-9-]{2,5})+";

	// Generate pattern
	$regex = "^".$first."@".$domain."$";

	// Return check result
	return eregi($regex, $email);
}
// Function taken from user comments on www.php.net / function eregi()
function VALIDATE_URL ($URL, $compile=true) {
	// Trim URL a little
	$URL = trim(urldecode($URL));
	//* DEBUG: */ echo $URL."<br />";

	// Compile some chars out...
	if ($compile) $URL = COMPILE_CODE($URL, false, false, false);
	//* DEBUG: */ echo $URL."<br />";

	// Check for the extension filter
	if (EXT_IS_ACTIVE("filter")) {
		// Use the extension's filter set
		return FILTER_VALIDATE_URL($URL, false);
	}

	// If not installed, perform a simple test. Just make it sure there is always a http:// or
	// https:// in front of the URLs
	return (((substr($URL, 0, 7) == "http://") || (substr($URL, 0, 8) == "https://")) && (strlen($URL) >= 12));
}
//
function MEMBER_ACTION_LINKS($uid, $status="") {
	// Define all main targets
	$TARGETS = array("del_user", "edit_user", "lock_user", "add_points", "sub_points");

	// Begin of navigation links
	$eval = "\$OUT = \"[&nbsp;";

	foreach ($TARGETS as $tar) {
		$eval .= "<SPAN class=\\\"admin_user_link\\\"><A href=\\\"".URL."/modules.php?module=admin&amp;what=".$tar."&amp;u_id=".$uid."\\\" title=\\\"\".ADMIN_LINK_";
		//* DEBUG: */ echo "*".$tar."/".$status."*<br />\n";
		if (($tar == "lock_user") && ($status == "LOCKED")) {
			// Locked accounts shall be unlocked
			$eval .= "UNLOCK_USER";
		} else {
			// All other status is fine
			$eval .= strtoupper($tar);
		}
		$eval .= "_TITLE.\"\\\">\".ADMIN_";
		if (($tar == "lock_user") && ($status == "LOCKED")) {
			// Locked accounts shall be unlocked
			$eval .= "UNLOCK_USER";
		} else {
			// All other status is fine
			$eval .= strtoupper($tar);
		}
		$eval .= ".\"</A></SPAN>&nbsp;|&nbsp;";
	}

	// Finish navigation link
	$eval = substr($eval, 0, -7)."]\";";
	eval($eval);

	// Return string
	return $OUT;
}
// Function for backward-compatiblity
function ADD_CATEGORY_TABLE ($MODE, $return=false) {
	// Load it from the register extension
	return REGISTER_ADD_CATEGORY_TABLE ($MODE, $return);
}
// Generate an email link
function CREATE_EMAIL_LINK ($email, $table = "admins") {
	// Default email link (INSECURE! Spammer can read this by harvester programs)
	$EMAIL = "mailto:".$email;

	// Check for several extensions
	if ((EXT_IS_ACTIVE("admins")) && ($table == "admins")) {
		// Create email link for contacting admin in guest area
		$EMAIL = ADMINS_CREATE_EMAIL_LINK($email);
	} elseif ((EXT_IS_ACTIVE("user")) && (GET_EXT_VERSION("user") >= "0.3.3") && ($table == "user_data")) {
		// Create email link for contacting a member within admin area (or later in other areas, too?)
		$EMAIL = USER_CREATE_EMAIL_LINK($email);
	} elseif ((EXT_IS_ACTIVE("sponsor")) && ($table == "sponsor_data")) {
		// Create email link to contact sponsor within admin area (or like the link above?)
		$EMAIL = SPONSOR_CREATE_EMAIL_LINK($email);
	}

	// Shall I close the link when there is no admin?
	if ((!IS_ADMIN()) && ($EMAIL == $email)) $EMAIL = "#"; // Closed!

	// Return email link
	return $EMAIL;
}
// Generate a hash for extra-security for all passwords
function generateHash ($plainText, $salt = "") {
	global $_CONFIG, $_SERVER;

	// Is the required extension "sql_patches" there and a salt is not given?
	if (((EXT_VERSION_IS_OLDER("sql_patches", "0.3.6")) || (!EXT_IS_ACTIVE("sql_patches"))) && (empty($salt))) {
		// Extension sql_patches is missing/outdated so we hash the plain text with MD5
		return md5($plainText);
	} // END - if

	// Do we miss an arry element here?
	if (getConfig('file_hash') == null) {
		// Stop here
		print("Missing file_hash in ".__FUNCTION__.". Backtrace:<pre>");
		debug_print_backtrace();
		die("</pre>");
	} // END - if

	// When the salt is empty build a new one, else use the first x configured characters as the salt
	if (empty($salt)) {
		// Build server string
		$server = $_SERVER['PHP_SELF'].":".GET_USER_AGENT().":".getenv('SERVER_SOFTWARE').":".GET_REMOTE_ADDR().":".":".filemtime(PATH."inc/databases.php");

		// Build key string
		$keys   = SITE_KEY.":".DATE_KEY.":".getConfig('secret_key').":".getConfig('file_hash').":".date("d-m-Y (l-F-T)", bigintval(getConfig('patch_ctime'))).":".getConfig('master_salt');

		// Additional data
		$data = $plainText.":".uniqid(mt_rand(), true).":".time();

		// Calculate number for generating the code
		$a = time() + _ADD - 1;

		// Generate SHA1 sum from modula of number and the prime number
		$sha1 = sha1(($a % _PRIME).$server.":".$keys.":".$data.":".date("d-m-Y (l-F-T)", time()).":".$a);
		//* DEBUG: */ echo "SHA1=".$sha1." (".strlen($sha1).")<br />";
		$sha1 = scrambleString($sha1);
		//* DEBUG: */ echo "Scrambled=".$sha1." (".strlen($sha1).")<br />";
		//* DEBUG: */ $sha1b = descrambleString($sha1);
		//* DEBUG: */ echo "Descrambled=".$sha1b." (".strlen($sha1b).")<br />";

		// Generate the password salt string
		$salt = substr($sha1, 0, getConfig('salt_length'));
		//* DEBUG: */ echo $salt." (".strlen($salt).")<br />";
	} else {
		// Use given salt
		$salt = substr($salt, 0, getConfig('salt_length'));
		//* DEBUG: */ echo "GIVEN={$salt}<br />\n";
	}

	// Return hash
	return $salt.sha1($salt.$plainText);
}
//
function scrambleString($str) {
	global $_CONFIG;

	// Init
	$scrambled = "";

	// Final check, in case of failture it will return unscrambled string
	if (strlen($str) > 40) {
		// The string is to long
		return $str;
	} elseif (strlen($str) == 40) {
		// From database
		$scrambleNums = explode(":", getConfig('pass_scramble'));
	} else {
		// Generate new numbers
		$scrambleNums = explode(":", genScrambleString(strlen($str)));
	}

	// Scramble string here
	//* DEBUG: */ echo "***Original=".$str."***<br />";
	for ($idx = 0; $idx < strlen($str); $idx++) {
		// Get char on scrambled position
		$char = substr($str, $scrambleNums[$idx], 1);

		// Add it to final output string
		$scrambled .= $char;
	} // END - for

	// Return scrambled string
	//* DEBUG: */ echo "***Scrambled=".$scrambled."***<br />";
	return $scrambled;
}
//
function descrambleString($str) {
	global $_CONFIG;
	// Scramble only 40 chars long strings
	if (strlen($str) != 40) return $str;

	// Load numbers from config
	$scrambleNums = explode(":", getConfig('pass_scramble'));

	// Validate numbers
	if (count($scrambleNums) != 40) return $str;

	// Begin descrambling
	$orig = str_repeat(" ", 40);
	//* DEBUG: */ echo "+++Scrambled=".$str."+++<br />";
	for ($idx = 0; $idx < 40; $idx++) {
		$char = substr($str, $idx, 1);
		$orig = substr_replace($orig, $char, $scrambleNums[$idx], 1);
	} // END - for

	// Return scrambled string
	//* DEBUG: */ echo "+++Original=".$orig."+++<br />";
	return $orig;
}
//
function genScrambleString($len) {
	// Prepare randomizer and array for the numbers
	mt_srand((double) microtime() * 1000000);
	$scrambleNumbers = array();

	// First we need to setup randomized numbers from 0 to 31
	for ($idx = 0; $idx < $len; $idx++) {
		// Generate number
		$rand = mt_rand(0, ($len -1));

		// Check for it by creating more numbers
		while (array_key_exists($rand, $scrambleNumbers)) {
			$rand = mt_rand(0, ($len -1));
		} // END - while

		// Add number
		$scrambleNumbers[$rand] = $rand;
	} // END - for

	// So let's create the string for storing it in database
	$scrambleString = implode(":", $scrambleNumbers);
	return $scrambleString;
}
// Append data like session ID or referal ID to the given URL which would
// normally be stored in cookies
function ADD_URL_DATA($URL) {
	global $_CONFIG;
	$ADD = "";

	// Determine URL binder
	$BIND = "?";
	if (strpos($URL, "?") !== false) $BIND = "&";

	if ((!defined('__COOKIES')) || ((!__COOKIES))) {
		// Cookies are not accepted
		if ((!empty($_GET['refid'])) && (strpos($URL, "refid=") == 0)) {
			// Cookie found in URL
			$ADD .= $BIND."refid=".bigintval($_GET['refid']);
		} elseif ((GET_EXT_VERSION("sql_patches") != '') && (getConfig('def_refid') > 0)) {
			// Not found! So let's set default here
			$ADD .= $BIND."refid=".getConfig('def_refid');
		}

		// Is there already added data? Then change the binder
		if (!empty($ADD)) $BIND = "&";

		// Add session ID
		if ((!empty($_GET['PHPSESSID'])) && (strpos($URL, "PHPSESSID=") == 0)) {
			// Add session from URL
			$ADD .= $BIND."PHPSESSID=".SQL_ESCAPE(strip_tags($_GET['PHPSESSID']));
		} else {
			// Add current session
			$ADD .= $BIND."PHPSESSID=".session_id();
		}
	} // END - if

	// Add all together and return it
	return $URL.$ADD;
}
// Generate an PGP-like encrypted hash of given hash for e.g. cookies
function generatePassString($passHash) {
	global $_CONFIG;

	// Return vanilla password hash
	$ret = $passHash;

	// Is a secret key and master salt already initialized?
	if ((getConfig('secret_key') != "") && (getConfig('master_salt') != "")) {
		// Only calculate when the secret key is generated
		$newHash = ""; $start = 9;
		for ($idx = 0; $idx < 10; $idx++) {
			$part1 = hexdec(substr($passHash, $start, 4));
			$part2 = hexdec(substr(getConfig('secret_key'), $start, 4));
			$mod = dechex($idx);
			if ($part1 > $part2) {
				$mod = dechex(sqrt(($part1 - $part2) * _PRIME / pi()));
			} elseif ($part2 > $part1) {
				$mod = dechex(sqrt(($part2 - $part1) * _PRIME / pi()));
			}
			$mod = substr(round($mod), 0, 4);
			$mod = str_repeat('0', 4-strlen($mod)).$mod;
			//* DEBUG: */ echo "*".$start."=".$mod."*<br />";
			$start += 4;
			$newHash .= $mod;
		} // END - for

		//* DEBUG: */ print($passHash."<br />".$newHash." (".strlen($newHash).")");
		$ret = generateHash($newHash, getConfig('master_salt'));
		//* DEBUG: */ print($ret."<br />\n");
	} else {
		// Hash it simple
		//* DEBUG: */ echo "--".$passHash."--<br />\n";
		$ret = md5($passHash);
		//* DEBUG: */ echo "++".$ret."++<br />\n";
	}

	// Return result
	return $ret;
}

// Fix "deleted" cookies
function FIX_DELETED_COOKIES ($cookies) {
	// Is this an array with entries?
	if ((is_array($cookies)) && (count($cookies) > 0)) {
		// Then check all cookies if they are marked as deleted!
		foreach ($cookies as $cookieName) {
			// Is the cookie set to "deleted"?
			if (get_session($cookieName) == "deleted") {
				set_session($cookieName, "");
			}
		} // END - foreach
	} // END - if
}

// Output error messages in a fasioned way and die...
function mxchange_die ($msg) {
	global $footer;

	// Load the message template
	LOAD_TEMPLATE("admin_settings_saved", false, $msg);

	// Load footer
	include(PATH."inc/footer.php");

	// Exit explicitly
	exit;
}

// Display parsing time and number of SQL queries in footer
function DISPLAY_PARSING_TIME_FOOTER() {
	global $_CONFIG;

	// Is the timer started?
	if (!isset($GLOBALS['startTime'])) {
		// Abort here
		return false;
	} // END - if

	// Get end time
	$endTime = microtime(true);

	// "Explode" both times
	$start = explode(" ", $GLOBALS['startTime']);
	$end = explode(" ", $endTime);
	$runTime = $end[0] - $start[0];
	if ($runTime < 0) $runTime = 0;
	$runTime = TRANSLATE_COMMA($runTime);

	// Prepare output
	$content = array(
		'runtime' 		=> $runTime,
		'numSQLs'		=> (getConfig('sql_count') + 1),
		'numTemplates'	=> (getConfig('num_templates') + 1)
	);

	// Load the template
	LOAD_TEMPLATE("show_timings", false, $content);
}

// Unset/set session variables
function set_session ($var, $value) {
	global $CSS;

	// Abort in CSS mode here
	if ($CSS == 1) return true;

	// Trim value and session variable
	$var = trim(SQL_ESCAPE($var)); $value = trim($value);

	// Is the session variable set?
	if (("".$value."" == "") && (isSessionVariableSet($var))) {
		// Remove the session
		//* DEBUG: */ echo "UNSET:".$var."=".get_session($var)."<br />\n";
		unset($_SESSION[$var]);
		return session_unregister($var);
	} elseif (("".$value."" != '') && (!isSessionVariableSet($var))) {
		// Set session
		//* DEBUG: */ echo "SET:".$var."=".$value."<br />\n";
		$_SESSION[$var] =  $value;
		return session_register($var);
	} elseif (!empty($value)) {
		// Update session
		//* DEBUG: */ echo "UPDATE:".$var."=".$value."<br />\n";
		$_SESSION[$var] = $value;
		return true;
	}

	// Ignored (but valid)
	//* DEBUG: */ echo "IGNORED:".$var."=".$value."<br />\n";
	return true;
}

// Check wether a boolean constant is set
// Taken from user comments in PHP documentation for function constant()
function isBooleanConstantAndTrue($constName) { // : Boolean
	global $cacheArray;

	// Failed by default
	$res = false;

	// In cache?
	if (isset($cacheArray['const'][$constName])) {
		// Use cache
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ".$constName."-CACHE!<br />\n";
		$res = $cacheArray['const'][$constName];
	} else {
		// Check constant
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ".$constName."-RESOLVE!<br />\n";
		if (defined($constName)) $res = (constant($constName) === true);

		// Set cache
		$cacheArray['const'][$constName] = $res;
	}
	//* DEBUG: */ var_dump($res);

	// Return value
	return $res;
}

// Check wether a session variable is set
function isSessionVariableSet ($var) {
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):var={$var}<br />\n";
	return (isset($_SESSION[$var]));
}
// Returns wether the value of the session variable or NULL if not set
function get_session ($var) {
	global $cacheArray;

	// Default is not found! ;-)
	$value = null;

	// Is the variable there or cached values?
	if (isset($cacheArray['session'][$var])) {
		// Get cached value (skips a lot SQL_ESCAPE() calles!
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ".$var."-CACHE!<br />\n";
		$value = $cacheArray['session'][$var];
	} elseif (isSessionVariableSet($var)) {
		// Then  get it secured!
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ".$var."-RESOLVE!<br />\n";
		$value = SQL_ESCAPE($_SESSION[$var]);

		// Cache the value
		$cacheArray['session'][$var] = $value;
	} // END - if

	// Return the value
	return $value;
}
// Send notification to admin
function SEND_ADMIN_NOTIFICATION($subject, $templateName, $content=array(), $uid="0") {
	if (GET_EXT_VERSION("admins") >= "0.4.1") {
		// Send new way
		SEND_ADMIN_EMAILS_PRO($subject, $templateName, $content, $uid);
	} else {
		// Send outdated way
		$msg = LOAD_EMAIL_TEMPLATE($templateName, $content, $uid);
		SEND_ADMIN_EMAILS($subject, $msg);
	}
}
// Destroy user session
function destroy_user_session () {
	// Remove all user data from session
	return ((set_session("userid", "")) && (set_session("u_hash", "")) && (set_session("lifetime", "")));
}
// Merges an array together but only if both are arrays
function merge_array ($array1, $array2) {
	// Are both an array?
	if ((is_array($array1)) && (is_array($array2))) {
		// Merge all together
		return array_merge($array1, $array2);
	} elseif (is_array($array1)) {
		// Return left array
		return $array1;
	}

	// Something wired happened here...
	print(__FUNCTION__.":<pre>");
	debug_print_backtrace();
	die("</pre>");
}
// Debug message logger
function DEBUG_LOG ($file, $line, $message, $force=true) {
	// Is debug mode enabled?
	if ((isBooleanConstantAndTrue('DEBUG_MODE')) || ($force)) {
		// Log this message away
		$fp = fopen(PATH."inc/cache/debug.log", 'a') or mxchange_die("Cannot write logfile debug.log!");
		fwrite($fp, date("d.m.Y|H:i:s", time())."|".basename($file)."|".$line."|".strip_tags($message)."\n");
		fclose($fp);
	} // END - if
}
// Reads a directory with PHP files in and gets only files back
function GET_DIR_AS_ARRAY ($baseDir, $prefix) {
	$INCs = array();

	// Open directory
	$dirPointer = opendir($baseDir) or mxchange_die("Cannot read ".basename($baseDir)." path!");

	// Read all entries
	while ($baseFile = readdir($dirPointer)) {
		// Load file only if extension is active
		// Make full path
		$file = $baseDir.$baseFile;

		// Is this a valid reset file?
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):baseDir={$baseDir},prefix={$prefix},baseFile={$baseFile}<br />\n";
		if ((FILE_READABLE($file)) && (substr($baseFile, 0, strlen($prefix)) == $prefix) && (substr($baseFile, -4, 4) == ".php")) {
			// Remove both for extension name
			$extName = substr($baseFile, strlen($prefix), -4);

			// Try to find it
			$extId = GET_EXT_ID($extName);

			// Is the extension valid and active?
			if (($extId > 0) && (EXT_IS_ACTIVE($extName))) {
				// Then add this file
				$INCs[] = $file;
			} elseif ($extId == 0) {
				// Add non-extension files as well
				$INCs[] = $file;
			}
		} // END - if
	} // END - while

	// Close directory
	closedir($dirPointer);

	// Sort array
	asort($INCs);

	// Return array with include files
	return $INCs;
}
// Load more reset scripts
function RESET_ADD_INCLUDES () {
	global $_CONFIG;

	// Is the reset set or old sql_patches?
	if ((!defined('__DAILY_RESET')) || (EXT_VERSION_IS_OLDER("sql_patches", "0.4.5"))) {
		// Then abort here
		return array();
	} // END - if

	// Get more daily reset scripts
	$INC_POOL = GET_DIR_AS_ARRAY(PATH."inc/reset/", "reset_");

	// Update database
	if (!defined('DEBUG_RESET')) UPDATE_CONFIG("last_update", time());

	// Create current week mark
	$currWeek = date("W", time());

	// Has it changed?
	if (getConfig('last_week') != $currWeek) {
		// Include weekly reset scripts
		$INC_POOL = merge_array($INC_POOL, GET_DIR_AS_ARRAY(PATH."inc/weekly/", "weekly_"));

		// Update config
		if (!defined('DEBUG_WEEKLY')) UPDATE_CONFIG("last_week", $currWeek);
	} // END - if

	// Create current month mark
	$currMonth = date("m", time());

	// Has it changed?
	if (getConfig('last_month') != $currMonth) {
		// Include monthly reset scripts
		$INC_POOL = merge_array($INC_POOL, GET_DIR_AS_ARRAY(PATH."inc/monthly/", "monthly_"));

		// Update config
		if (!defined('DEBUG_MONTHLY')) UPDATE_CONFIG("last_month", $currMonth);
	} // END - if

	// Return array
	return $INC_POOL;
}
// Handle extra values
function HANDLE_EXTRA_VALUES ($filterFunction, $value, $extraValue) {
	// Default is the value itself
	$ret = $value;

	// Do we have a special filter function?
	if (!empty($filterFunction)) {
		// Does the filter function exist?
		if (function_exists($filterFunction)) {
			// Do we have extra parameters here?
			if (!empty($extraValue)) {
				// Put both parameters in one new array by default
				$args = array($value, $extraValue);

				// If we have an array simply use it and pre-extend it with our value
				if (is_array($extraValue)) {
					// Make the new args array
					$args = merge_array(array($value), $extraValue);
				} // END - if

				// Call the multi-parameter call-back
				$ret = call_user_func_array($filterFunction, $args);
			} else {
				// One parameter call
				$ret = call_user_func($filterFunction, $value);
			}
		} // END - if
	} // END - if

	// Return the value
	return $ret;
}
// Check if given FQFN is a readable file
function FILE_READABLE($fqfn) {
	// Check all...
	return ((file_exists($fqfn)) && (is_file($fqfn)) && (is_readable($fqfn)));
}
// Converts timestamp selections into a timestamp
function CONVERT_SELECTIONS_TO_TIMESTAMP(&$POST, &$DATA, &$id, &$skip) {
	// Init test variable
	$test2 = "";

	// Get last three chars
	$test = substr($id, -3);

	// Improved way of checking! :-)
	if (in_array($test, array("_ye", "_mo", "_we", "_da", "_ho", "_mi", "_se"))) {
		// Found a multi-selection for timings?
		$test = substr($id, 0, -3);
		if ((isset($POST[$test."_ye"])) && (isset($POST[$test."_mo"])) && (isset($POST[$test."_we"])) && (isset($POST[$test."_da"])) && (isset($POST[$test."_ho"])) && (isset($POST[$test."_mi"])) && (isset($POST[$test."_se"])) && ($test != $test2)) {
			// Generate timestamp
			$POST[$test] = CREATE_TIMESTAMP_FROM_SELECTIONS($test, $POST);
			$DATA[] = sprintf("%s='%s'", $test, $POST[$test]);

			// Remove data from array
			foreach (array("ye", "mo", "we", "da", "ho", "mi", "se") as $rem) {
				unset($POST[$test."_".$rem]);
			} // END - foreach

			// Skip adding
			unset($id); $skip = true; $test2 = $test;
		} // END - if
	} else {
		// Process this entry
		$skip = false; $test2 = "";
	}
}
// Reverts the german decimal comma into Computer decimal dot
function REVERT_COMMA ($str) {
	// Default float is not a float... ;-)
	$float = false;

	// Which language is selected?
	switch (GET_LANGUAGE()) {
		case "de": // German language
			// Remove german thousand dots first
			$str = str_replace(".", "", $str);

			// Replace german commata with decimal dot and cast it
			$float = (float)str_replace(",", ".", $str);
			break;

		default: // US and so on
			// Remove thousand dots first and cast
			$float = (float)str_replace(",", "", $str);
			break;
	}

	// Return float
	return $float;
}
// Handle menu-depending failed logins and return the rendered content
function HANDLE_LOGIN_FAILTURES ($accessLevel) {
	// Default output is empty ;-)
	$OUT = "";

	// Is the session data set?
	if ((isSessionVariableSet('mxchange_'.$accessLevel.'_failtures')) && (isSessionVariableSet('mxchange_'.$accessLevel.'_last_fail'))) {
		// Ignore zero values
		if (get_session('mxchange_'.$accessLevel.'_failtures') > 0) {
			// Non-guest has login failtures found, get both data and prepare it for template
			//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):accessLevel={$accessLevel}<br />\n";
			$content = array(
				'login_failtures' => get_session('mxchange_'.$accessLevel.'_failtures'),
				'last_failture'   => MAKE_DATETIME(get_session('mxchange_'.$accessLevel.'_last_fail'), "2")
			);

			// Load template
			$OUT = LOAD_TEMPLATE("login_failtures", true, $content);
		} // END - if

		// Reset session data
		set_session('mxchange_'.$accessLevel.'_failtures', "");
		set_session('mxchange_'.$accessLevel.'_last_fail', "");
	} // END - if

	// Return rendered content
	return $OUT;
}
// Rebuild cache
function REBUILD_CACHE ($cache, $inc="") {
	global $cacheInstance, $_CONFIG, $CSS;

	// Shall I remove the cache file?
	if ((EXT_IS_ACTIVE("cache")) && (is_object($cacheInstance))) {
		// Rebuild cache
		if ($cacheInstance->loadCacheFile($cache)) {
			// Destroy it
			$cacheInstance->destroyCacheFile();
		} // END - if

		// Include file given?
		if (!empty($inc)) {
			// Construct FQFN
			$fqfn = sprintf("%sinc/loader/load_cache-%s.php", PATH, $inc);

			// Is the include there?
			if (FILE_READABLE($fqfn)) {
				// And rebuild it from scratch
				//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): inc={$inc} - LOADED!<br />\n";
				require($fqfn);
			} else {
				// Include not found!
				DEBUG_LOG(__FUNCTION__, __LINE__, "Include {$inc} not found. cache={$cache}");
			}
		} // END - if
	} // END - if
}
// Purge admin menu cache
function CACHE_PURGE_ADMIN_MENU ($id=0, $action="", $what="", $str="") {
	global $cacheInstance;

	// Is the cache extension enabled or no cache instance or admin menu cache disabled?
	if (!EXT_IS_ACTIVE("cache")) {
		// Cache extension not active
		return false;
	} elseif (!is_object($cacheInstance)) {
		// No cache instance!
		DEBUG_LOG(__FUNCTION__, __LINE__, " No cache instance found.");
		return false;
	} elseif ((getConfig('cache_admin_menu') == null) || (getConfig('cache_admin_menu') == "N")) {
		// Caching disabled (currently experiemental!)
		return false;
	}

	// Experiemental feature!
	trigger_error("You have to delete the admin_*.cache files by yourself at this point.");
}
// Translates the "pool type" into human-readable
function TRANSLATE_POOL_TYPE ($type) {
	// Default type is unknown
	$translated = sprintf(POOL_TYPE_UNKNOWN, $type);

	// Generate constant
	$constName = sprintf("POOL_TYPE_%s", $type);

	// Does it exist?
	if (defined($constName)) {
		// Then use it
		$translated = constant($constName);
	} // END - if

	// Return "translation"
	return $translated;
}
// "Getter" for remote IP number
function GET_REMOTE_ADDR () {
	// Get remote ip from environment
	$remoteAddr = getenv('REMOTE_ADDR');

	// Is removeip installed?
	if (EXT_IS_ACTIVE("removeip")) {
		// Then anonymize it
		$remoteAddr = GET_ANONYMOUS_REMOTE_ADDR($remoteAddr);
	} // END - if

	// Return it
	return $remoteAddr;
}
// "Getter" for remote hostname
function GET_REMOTE_HOST () {
	// Get remote ip from environment
	$remoteHost = getenv('REMOTE_HOST');

	// Is removeip installed?
	if (EXT_IS_ACTIVE("removeip")) {
		// Then anonymize it
		$remoteHost = GET_ANONYMOUS_REMOTE_HOST($remoteHost);
	} // END - if

	// Return it
	return $remoteHost;
}
// "Getter" for user agent
function GET_USER_AGENT () {
	// Get remote ip from environment
	$userAgent = getenv('HTTP_USER_AGENT');

	// Is removeip installed?
	if (EXT_IS_ACTIVE("removeip")) {
		// Then anonymize it
		$userAgent = GET_ANONYMOUS_USER_AGENT($userAgent);
	} // END - if

	// Return it
	return $userAgent;
}
// "Getter" for referer
function GET_REFERER () {
	// Get remote ip from environment
	$referer = getenv('HTTP_REFERER');

	// Is removeip installed?
	if (EXT_IS_ACTIVE("removeip")) {
		// Then anonymize it
		$referer = GET_ANONYMOUS_REFERER($referer);
	} // END - if

	// Return it
	return $referer;
}

// Adds a bonus mail to the queue
// This is a high-level function!
function ADD_NEW_BONUS_MAIL ($data, $mode="", $output=true) {
	// Use mode from data if not set and availble ;-)
	if ((empty($mode)) && (isset($data['mode']))) $mode = $data['mode'];

	// Generate receiver list
	$RECEIVER = GENERATE_RECEIVER_LIST($data['cat'], $data['receiver'], $mode);

	// Receivers added?
	if (!empty($RECEIVER)) {
		// Add bonus mail to queue
		ADD_BONUS_MAIL_TO_QUEUE(
			$data['subject'],
			$data['text'],
			$RECEIVER,
			$data['points'],
			$data['seconds'],
			$data['url'],
			$data['cat'],
			$mode,
			$data['receiver']
		);

		// Mail inserted into bonus pool
		if ($output) LOAD_TEMPLATE("admin_settings_saved", false, ADMIN_BONUS_SEND);
	} elseif ($output) {
		// More entered than can be reached!
		LOAD_TEMPLATE("admin_settings_saved", false, ADMIN_MORE_SELECTED);
	} else {
		// Debug log
		DEBUG_LOG(__FUNCTION__, __LINE__, " cat={$data['cat']},receiver={$data['receiver']},data=".base64_encode(serialize($data))." More selected, than available!");
	}
}
// Determines referal id and sets it
function DETERMINE_REFID () {
	global $_CONFIG, $CLICK, $_SERVER;

	// Check if refid is set
	if ((!empty($_GET['user'])) && ($CLICK == 1) && (basename($_SERVER['PHP_SELF']) == "click.php")) {
		// The variable user comes from the click-counter script click.php and we only accept this here
		$GLOBALS['refid'] = bigintval($_GET['user']);
	} elseif (!empty($_POST['refid'])) {
		// Get referal id from variable refid (so I hope this makes my script more compatible to other scripts)
		$GLOBALS['refid'] = SQL_ESCAPE(strip_tags($_POST['refid']));
	} elseif (!empty($_GET['refid'])) {
		// Get referal id from variable refid (so I hope this makes my script more compatible to other scripts)
		$GLOBALS['refid'] = SQL_ESCAPE(strip_tags($_GET['refid']));
	} elseif (!empty($_GET['ref'])) {
		// Set refid=ref (the referal link uses such variable)
		$GLOBALS['refid'] = SQL_ESCAPE(strip_tags($_GET['ref']));
	} elseif ((isSessionVariableSet('refid')) && (get_session('refid') != 0)) {
		// Set session refid als global
		$GLOBALS['refid'] = bigintval(get_session('refid'));
	} elseif ((GET_EXT_VERSION("sql_patches") != "") && (getConfig('def_refid') > 0)) {
		// Set default refid as refid in URL
		$GLOBALS['refid'] = bigintval(getConfig('def_refid'));
	} elseif ((GET_EXT_VERSION("user") >= "0.3.4") && (getConfig('select_user_zero_refid')) == "Y") {
		// Select a random user which has confirmed enougth mails
		$GLOBALS['refid'] = SELECT_RANDOM_REFID();
	} else {
		// No default ID when sql_patches is not installed or none set
		$GLOBALS['refid'] = 0;
	}

	// Set cookie when default refid > 0
	if (!isSessionVariableSet('refid') || (!empty($GLOBALS['refid'])) || ((get_session('refid') == "0") && (getConfig('def_refid') > 0))) {
		// Set cookie
		set_session('refid', $GLOBALS['refid']);
	} // END - if

	// Return determined refid
	return $GLOBALS['refid'];
}

// Destroys the admin session
function DESTROY_ADMIN_SESSION ($destroy = true) {
	// Kill maybe existing session variables including array elements
	set_session('admin_login', "");
	set_session('admin_md5'  , "");
	set_session('admin_last' , "");
	set_session('admin_to'   , "");

	// Destroy session and return status
	if ($destroy) {
		return @session_destroy();
	} // END - if

	// All fine if we shall not really destroy the session
	return true;
}

// Checks if a given apache module is loaded
function IF_APACHE_MODULE_LOADED ($apacheModule) {
	// Check it and return result
	return (((function_exists('apache_get_modules')) && (in_array($apacheModule, apache_get_modules()))) || (!function_exists('apache_get_modules')));
}

// Getter for $_CONFIG entries
function getConfig ($entry) {
	global $_CONFIG;

	// Default value
	$value = null;

	// Is the entry there?
	if (isset($_CONFIG[$entry])) {
		// Then use it
		$value = $_CONFIG[$entry];
	} // END - if

	// Return it
	return $value;
}

// @TODO Rewrite all language constants to the function getLanguage().
// "Getter" for language strings
function getMessage ($messageId) {
	// Default is not found!
	$return = "!".$messageId."!";

	// Is the language string found?
	if (isset($GLOBALS['msg'][$messageId])) {
		// Language array element found
		$return = $GLOBALS['msg'][$messageId];
	} elseif (defined($messageId)) {
		// @DEPRECATED Deprecated constant found
		$return = constant($messageId);
	} else {
		// Missing language constant
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Missing message string %s detected.", $messageId));
	}

	// Return the string
	return $return;
}

// Get current theme name
function GET_CURR_THEME() {
	global $INC_POOL, $_CONFIG, $CSS, $cacheArray;

	// The default theme is 'default'... ;-)
	$ret = "default";

	// Load default theme if not empty from configuration
	if (getConfig('default_theme') != "") $ret = getConfig('default_theme');

	if (!isSessionVariableSet('mxchange_theme')) {
		// Set default theme
		set_session("mxchange_theme", $ret);
	} elseif ((isSessionVariableSet('mxchange_theme')) && (GET_EXT_VERSION("sql_patches") >= "0.1.4")) {
		//die("<pre>".print_r($cacheArray['themes'], true)."</pre>");
		// Get theme from cookie
		$ret = get_session('mxchange_theme');

		// Is it valid?
		if (THEME_GET_ID($ret) == 0) {
			// Fix it to default
			$ret = "default";
		} // END - if
	} elseif ((!isBooleanConstantAndTrue('mxchange_installed')) && ((isBooleanConstantAndTrue('mxchange_installing')) || ($CSS == true)) && ((!empty($_GET['theme'])) || (!empty($_POST['theme'])))) {
		// Prepare FQFN for checking
		$theme = sprintf("%stheme/%s/theme.php", PATH, SQL_ESCAPE($_GET['theme']));

		// Installation mode active
		if ((!empty($_GET['theme'])) && (FILE_READABLE($theme))) {
			// Set cookie from URL data
			set_session("mxchange_theme", SQL_ESCAPE($_GET['theme']));
		} elseif (FILE_READABLE(sprintf("%stheme/%s/theme.php", PATH, SQL_ESCAPE($_POST['theme'])))) {
			// Set cookie from posted data
			set_session("mxchange_theme", SQL_ESCAPE($_POST['theme']));
		}

		// Set return value
		$ret = get_session('mxchange_theme');
	} else {
		// Invalid design, reset cookie
		set_session("mxchange_theme", $ret);
	}

	// Add (maybe) found theme.php file to inclusion list
	$theme = sprintf("%stheme/%s/theme.php", PATH, SQL_ESCAPE($ret));

	// Try to load the requested include file
	if (FILE_READABLE($theme)) $INC_POOL[] = $theme;

	// Return theme value
	return $ret;
}

// Get id from theme
function THEME_GET_ID ($name) {
	global $cacheArray, $_CONFIG;

	// Is the extension "theme" installed?
	if (!EXT_IS_ACTIVE("theme")) {
		// Then abort here
		return 0;
	} // END - if

	// Default id
	$id = 0;

	// Is the cache entry there?
	if (isset($cacheArray['themes']['id'][$name])) {
		// Get the version from cache
		$id = $cacheArray['themes']['id'][$name];

		// Count up
		if (getConfig('cache_hits') > 0) { $_CONFIG['cache_hits']++; } else { $_CONFIG['cache_hits'] = 1; }
	} elseif (GET_EXT_VERSION("cache") != "0.1.8") {
		// Check if current theme is already imported or not
		$result = SQL_QUERY_ESC("SELECT id FROM `"._MYSQL_PREFIX."_themes` WHERE theme_path='%s' LIMIT 1",
			array($name), __FILE__, __LINE__);

		// Entry found?
		if (SQL_NUMROWS($result) == 1) {
			// Fetch data
			list($id) = SQL_FETCHROW($result);
		} // END - if

		// Free result
		SQL_FREERESULT($result);
	}

	// Return id
	return $id;
}

//////////////////////////////////////////////////
//                                              //
// AUTOMATICALLY RE-GENERATED MISSING FUNCTIONS //
//                                              //
//////////////////////////////////////////////////
//
if (!function_exists('html_entity_decode')) {
	// Taken from documentation on www.php.net
	function html_entity_decode ($string) {
		$trans_tbl = get_html_translation_table(HTML_ENTITIES);
		$trans_tbl = array_flip($trans_tbl);
		return strtr($string, $trans_tbl);
	}
} // END - if

//
?>
